# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
from datetime import datetime, timedelta
from os.path import isdir, isfile
from typing import Any, Dict, List, Optional, Set

import modin.pandas as modin_pd
import pandas as pd

from taipy.config.common.scope import Scope

from .._backup._backup import _replace_in_backup_file
from .._entity._reload import _self_reload
from .._version._version_manager_factory import _VersionManagerFactory
from ..exceptions.exceptions import InvalidExposedType, UnknownCompressionAlgorithm, UnknownParquetEngine
from ..job.job_id import JobId
from .abstract_file import _AbstractFileDataNode
from .data_node import DataNode
from .data_node_id import DataNodeId, Edit


class ParquetDataNode(DataNode, _AbstractFileDataNode):
    """Data Node stored as a Parquet file.

    Attributes:
        config_id (str): Identifier of the data node configuration. This string must be a valid
            Python identifier.
        scope (Scope^): The scope of this data node.
        id (str): The unique identifier of this data node.
        name (str): A user-readable name of this data node.
        owner_id (str): The identifier of the owner (pipeline_id, scenario_id, cycle_id) or `None`.
        parent_ids (Optional[Set[str]]): The identifiers of the parent tasks or `None`.
        last_edit_date (datetime): The date and time of the last modification.
        edits (List[Edit^]): The ordered list of edits for that job.
        version (str): The string indicates the application version of the data node to instantiate. If not provided,
            the current version is used.
        validity_period (Optional[timedelta]): The duration implemented as a timedelta since the last edit date for
            which the data node can be considered up-to-date. Once the validity period has passed, the data node is
            considered stale and relevant tasks will run even if they are skippable (see the
            [Task management page](../core/entities/task-mgt.md) for more details).
            If _validity_period_ is set to `None`, the data node is always up-to-date.
        edit_in_progress (bool): True if a task computing the data node has been submitted
            and not completed yet. False otherwise.
        path (str): The path to the Parquet file.
        properties (dict[str, Any]): A dictionary of additional properties. *properties*
            must have a *"default_path"* or *"path"* entry with the path of the Parquet file:

            - *"default_path"* (`str`): The default path of the Parquet file.
            - *"exposed_type"*: The exposed type of the data read from Parquet file. The default
                value is `pandas`.
            - *"engine"* (`Optional[str]`): Parquet library to use. Possible values are
                *"fastparquet"* or *"pyarrow"*.<br/>
                The default value is *"pyarrow"*.
            - *"compression"* (`Optional[str]`): Name of the compression to use. Possible values
                are *"snappy"*, *"gzip"*, *"brotli"*, or *"none"* (no compression).<br/>
                The default value is *"snappy"*.
            - *"read_kwargs"* (`Optional[dict]`): Additional parameters passed to the
                *pandas.read_parquet()* function.
            - *"write_kwargs"* (`Optional[dict]`): Additional parameters passed to the
                *pandas.DataFrame.write_parquet()* fucntion.
                The parameters in *"read_kwargs"* and *"write_kwargs"* have a
                **higher precedence** than the top-level parameters which are also passed to
                Pandas.
    """

    __STORAGE_TYPE = "parquet"
    __EXPOSED_TYPE_PROPERTY = "exposed_type"
    __EXPOSED_TYPE_NUMPY = "numpy"
    __EXPOSED_TYPE_PANDAS = "pandas"
    __EXPOSED_TYPE_MODIN = "modin"
    __VALID_STRING_EXPOSED_TYPES = [__EXPOSED_TYPE_PANDAS, __EXPOSED_TYPE_MODIN, __EXPOSED_TYPE_NUMPY]
    __PATH_KEY = "path"
    __DEFAULT_DATA_KEY = "default_data"
    __DEFAULT_PATH_KEY = "default_path"
    __ENGINE_PROPERTY = "engine"
    __VALID_PARQUET_ENGINES = ["pyarrow", "fastparquet"]
    __COMPRESSION_PROPERTY = "compression"
    __VALID_COMPRESSION_ALGORITHMS = ["snappy", "gzip", "brotli", "none"]
    __READ_KWARGS_PROPERTY = "read_kwargs"
    __WRITE_KWARGS_PROPERTY = "write_kwargs"
    _REQUIRED_PROPERTIES: List[str] = []

    def __init__(
        self,
        config_id: str,
        scope: Scope,
        id: Optional[DataNodeId] = None,
        name: Optional[str] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        last_edit_date: Optional[datetime] = None,
        edits: Optional[List[Edit]] = None,
        version: Optional[str] = None,
        validity_period: Optional[timedelta] = None,
        edit_in_progress: bool = False,
        properties: Optional[Dict] = None,
    ):
        if properties is None:
            properties = {}

        default_value = properties.pop(self.__DEFAULT_DATA_KEY, None)

        if self.__ENGINE_PROPERTY not in properties.keys():
            properties[self.__ENGINE_PROPERTY] = "pyarrow"
        if properties[self.__ENGINE_PROPERTY] not in self.__VALID_PARQUET_ENGINES:
            raise UnknownParquetEngine(
                f"Invalid parquet engine: {properties[self.__ENGINE_PROPERTY]}. "
                f"Supported engines are {', '.join(self.__VALID_PARQUET_ENGINES)}"
            )

        if self.__COMPRESSION_PROPERTY not in properties.keys():
            properties[self.__COMPRESSION_PROPERTY] = "snappy"
        if properties[self.__COMPRESSION_PROPERTY] == "none":
            properties[self.__COMPRESSION_PROPERTY] = None
        if (
            properties[self.__COMPRESSION_PROPERTY]
            and properties[self.__COMPRESSION_PROPERTY] not in self.__VALID_COMPRESSION_ALGORITHMS
        ):
            raise UnknownCompressionAlgorithm(
                f"Unsupported compression algorithm: {properties[self.__COMPRESSION_PROPERTY]}. "
                f"Supported algorithms are {', '.join(self.__VALID_COMPRESSION_ALGORITHMS)}"
            )

        if self.__READ_KWARGS_PROPERTY not in properties.keys():
            properties[self.__READ_KWARGS_PROPERTY] = {}

        if self.__WRITE_KWARGS_PROPERTY not in properties.keys():
            properties[self.__WRITE_KWARGS_PROPERTY] = {}

        if self.__EXPOSED_TYPE_PROPERTY not in properties.keys():
            properties[self.__EXPOSED_TYPE_PROPERTY] = self.__EXPOSED_TYPE_PANDAS
        self._check_exposed_type(properties[self.__EXPOSED_TYPE_PROPERTY])

        super().__init__(
            config_id,
            scope,
            id,
            name,
            owner_id,
            parent_ids,
            last_edit_date,
            edits,
            version or _VersionManagerFactory._build_manager()._get_latest_version(),
            validity_period,
            edit_in_progress,
            **properties,
        )
        self._path = properties.get(self.__PATH_KEY, properties.get(self.__DEFAULT_PATH_KEY))
        if not self._path:
            self._path = self._build_path(self.storage_type())
        properties[self.__PATH_KEY] = self._path

        if default_value is not None and not os.path.exists(self._path):
            self.write(default_value)

        if not self._last_edit_date and (isfile(self._path) or isdir(self._path)):
            self.last_edit_date = datetime.now()  # type: ignore

    @classmethod
    def storage_type(cls) -> str:
        return cls.__STORAGE_TYPE

    @property  # type: ignore
    @_self_reload(DataNode._MANAGER_NAME)
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        tmp_old_path = self._path
        self._path = value
        self.properties[self.__PATH_KEY] = value
        _replace_in_backup_file(old_file_path=tmp_old_path, new_file_path=self._path)

    def _check_exposed_type(self, exposed_type):
        if isinstance(exposed_type, str) and exposed_type not in self.__VALID_STRING_EXPOSED_TYPES:
            raise InvalidExposedType(
                f"Invalid string exposed type {exposed_type}. Supported values are "
                f"{', '.join(self.__VALID_STRING_EXPOSED_TYPES)}"
            )

    def _read(self):
        return self.read_with_kwargs()

    def _read_as(self, read_kwargs: Dict):
        custom_class = self.properties[self.__EXPOSED_TYPE_PROPERTY]
        list_of_dicts = self._read_as_pandas_dataframe(read_kwargs).to_dict(orient="records")
        return [custom_class(**dct) for dct in list_of_dicts]

    def _read_as_numpy(self, read_kwargs: Dict):
        return self._read_as_pandas_dataframe(read_kwargs).to_numpy()

    def _read_as_pandas_dataframe(self, read_kwargs: Dict) -> pd.DataFrame:
        return pd.read_parquet(self._path, **read_kwargs)

    def _read_as_modin_dataframe(self, read_kwargs: Dict) -> modin_pd.DataFrame:
        return modin_pd.read_parquet(self._path, **read_kwargs)

    def _write(self, data: Any):
        self.write_with_kwargs(data)

    def write_with_kwargs(self, data: Any, job_id: Optional[JobId] = None, **write_kwargs):
        """Write the data referenced by this data node.

        Keyword arguments here which are also present in the Data Node config will overwrite them.

        Parameters:
            data (Any): The data to write.
            job_id (JobId^): An optional identifier of the writer.
            **write_kwargs (dict[str, any]): The keyword arguments passed to the function
                `pandas.DataFrame.to_parquet()`.
        """
        kwargs = {
            self.__ENGINE_PROPERTY: self.properties[self.__ENGINE_PROPERTY],
            self.__COMPRESSION_PROPERTY: self.properties[self.__COMPRESSION_PROPERTY],
        }
        kwargs.update(self.properties[self.__WRITE_KWARGS_PROPERTY])
        kwargs.update(write_kwargs)
        if isinstance(data, (pd.DataFrame, modin_pd.DataFrame)):
            data.to_parquet(self._path, **kwargs)
        else:
            pd.DataFrame(data).to_parquet(self._path, **kwargs)
        self._track_edit(timestamp=datetime.now(), job_id=job_id)

    def read_with_kwargs(self, **read_kwargs):
        """Read data from this data node.

        Keyword arguments here which are also present in the Data Node config will overwrite them.

        Parameters:
            **read_kwargs (dict[str, any]): The keyword arguments passed to the function
                `pandas.read_parquet()`.
        """
        # return None if data was never written
        if not self.last_edit_date:
            self._DataNode__logger.warning(
                f"Data node {self.id} from config {self.config_id} is being read but has never been written."
            )
            return None

        kwargs = self.properties[self.__READ_KWARGS_PROPERTY]
        kwargs.update(
            {
                self.__ENGINE_PROPERTY: self.properties[self.__ENGINE_PROPERTY],
            }
        )
        kwargs.update(read_kwargs)

        if self.properties[self.__EXPOSED_TYPE_PROPERTY] == self.__EXPOSED_TYPE_PANDAS:
            return self._read_as_pandas_dataframe(kwargs)
        if self.properties[self.__EXPOSED_TYPE_PROPERTY] == self.__EXPOSED_TYPE_MODIN:
            return self._read_as_modin_dataframe(kwargs)
        if self.properties[self.__EXPOSED_TYPE_PROPERTY] == self.__EXPOSED_TYPE_NUMPY:
            return self._read_as_numpy(kwargs)
        return self._read_as(kwargs)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Any, Union

from taipy.config.config import Config

from .._repository._repository_factory import _RepositoryFactory
from ..common._utils import _load_fct
from ._scenario_fs_repository import _ScenarioFSRepository
from ._scenario_sql_repository import _ScenarioSQLRepository


class _ScenarioRepositoryFactory(_RepositoryFactory):
    _REPOSITORY_MAP = {"default": _ScenarioFSRepository, "sql": _ScenarioSQLRepository}

    @classmethod
    def _build_repository(cls) -> Union[_ScenarioFSRepository, Any]:  # type: ignore
        if cls._using_enterprise():
            factory = _load_fct(
                cls._TAIPY_ENTERPRISE_CORE_MODULE + ".scenario._scenario_repository_factory",
                "_ScenarioRepositoryFactory",
            )
            return factory._build_repository()  # type: ignore
        return cls._REPOSITORY_MAP.get(
            Config.global_config.repository_type, cls._REPOSITORY_MAP.get("default")
        )()  # type: ignore



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import pathlib
from typing import Any, Iterable, List, Optional, Union

from .._repository._repository import _AbstractRepository
from .._repository._repository_adapter import _RepositoryAdapter
from ..common._utils import _load_fct
from ..data._data_manager_factory import _DataManagerFactory
from ..exceptions.exceptions import NonExistingDataNode
from ._task_model import _TaskModel
from .task import Task
from .task_id import TaskId


class _TaskRepository(_AbstractRepository[_TaskModel, Task]):  # type: ignore
    def __init__(self, **kwargs):
        kwargs.update({"to_model_fct": self._to_model, "from_model_fct": self._from_model})
        self.repo = _RepositoryAdapter.select_base_repository()(**kwargs)

    @property
    def repository(self):
        return self.repo

    def _to_model(self, task: Task) -> _TaskModel:
        return _TaskModel(
            id=task.id,
            owner_id=task.owner_id,
            parent_ids=list(task._parent_ids),
            config_id=task.config_id,
            input_ids=self.__to_ids(task.input.values()),
            function_name=task._function.__name__,
            function_module=task._function.__module__,
            output_ids=self.__to_ids(task.output.values()),
            version=task.version,
            skippable=task._skippable,
            properties=task._properties.data.copy(),
        )

    def _from_model(self, model: _TaskModel) -> Task:
        return Task(
            id=TaskId(model.id),
            owner_id=model.owner_id,
            parent_ids=set(model.parent_ids),
            config_id=model.config_id,
            function=_load_fct(model.function_module, model.function_name),
            input=self.__to_data_nodes(model.input_ids),
            output=self.__to_data_nodes(model.output_ids),
            version=model.version,
            skippable=model.skippable,
            properties=model.properties,
        )

    def load(self, model_id: str) -> Task:
        return self.repo.load(model_id)

    def _load_all(self, version_number: Optional[str] = None) -> List[Task]:
        return self.repo._load_all(version_number)

    def _load_all_by(self, by, version_number: Optional[str] = None) -> List[Task]:
        return self.repo._load_all_by(by, version_number)

    def _save(self, entity: Task):
        return self.repo._save(entity)

    def _delete(self, entity_id: str):
        return self.repo._delete(entity_id)

    def _delete_all(self):
        return self.repo._delete_all()

    def _delete_many(self, ids: Iterable[str]):
        return self.repo._delete_many(ids)

    def _delete_by(self, attribute: str, value: str, version_number: Optional[str] = None):
        return self.repo._delete_by(attribute, value, version_number)

    def _search(self, attribute: str, value: Any, version_number: Optional[str] = None) -> Optional[Task]:
        return self.repo._search(attribute, value, version_number)

    def _export(self, entity_id: str, folder_path: Union[str, pathlib.Path]):
        return self.repo._export(entity_id, folder_path)

    @staticmethod
    def __to_ids(data_nodes):
        return [i.id for i in data_nodes]

    @staticmethod
    def __to_data_nodes(data_nodes_ids):
        data_nodes = []
        data_manager = _DataManagerFactory._build_manager()
        for _id in data_nodes_ids:
            if data_node := data_manager._get(_id):
                data_nodes.append(data_node)
            else:
                raise NonExistingDataNode(_id)
        return data_nodes

    def _get_by_config_id(self, config_id: str) -> List[Task]:
        return self.repo._get_by_config_id(config_id)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


from typing import Optional

from ._version_model import _VersionModel
from ._version_repository import _VersionRepository


class _VersionSQLRepository(_VersionRepository):
    _LATEST_VERSION_KEY = "latest_version"
    _DEVELOPMENT_VERSION_KEY = "development_version"
    _PRODUCTION_VERSION_KEY = "production_version"

    def __init__(self):
        super().__init__(model=_VersionModel, model_name="version")

    def _load_all_by(self, by, version_number: Optional[str] = "all"):
        return super()._load_all_by(by, version_number)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .._entity._reload import _get_manager


class _Entity:
    _MANAGER_NAME: str
    _is_in_context = False

    def __enter__(self):
        self._is_in_context = True
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._is_in_context = False
        _get_manager(self._MANAGER_NAME)._set(self)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import abc
from typing import Optional


class _Labeled:
    __LABEL_SEPARATOR = " > "

    @abc.abstractmethod
    def get_label(self) -> str:
        raise NotImplementedError

    def _get_label(self) -> str:
        """Returns the entity label made of the simple label prefixed by the owner label.

        Returns:
            The label of the entity as a string.
        """
        return self._get_explicit_label() or self._generate_label()

    @abc.abstractmethod
    def get_simple_label(self) -> str:
        raise NotImplementedError

    def _get_simple_label(self) -> str:
        """Returns the simple label.

        Returns:
            The simple label of the entity as a string.
        """
        return self._get_explicit_label() or self._generate_label(True)

    def _generate_label(self, simple=False) -> str:
        ls = []
        if not simple:
            if owner_id := self._get_owner_id():
                if getattr(self, "id") != owner_id:
                    from ... import core

                    owner = core.get(owner_id)
                    ls.append(owner.get_label())
        ls.append(self._generate_entity_label())
        return self.__LABEL_SEPARATOR.join(ls)

    def _get_explicit_label(self) -> Optional[str]:
        if hasattr(self, "_properties"):
            return getattr(self, "_properties").get("label")
        return None

    def _get_owner_id(self) -> Optional[str]:
        if hasattr(self, "owner_id"):
            return getattr(self, "owner_id")
        return None

    def _get_name(self) -> Optional[str]:
        if hasattr(self, "name"):
            return getattr(self, "name")
        if hasattr(self, "_properties"):
            return getattr(self, "_properties").get("name")
        return None

    def _get_config_id(self) -> Optional[str]:
        if hasattr(self, "config_id"):
            return getattr(self, "config_id")
        return None

    def _generate_entity_label(self) -> str:
        if name := self._get_name():
            return name
        if config_id := self._get_config_id():
            return config_id
        return getattr(self, "id")



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Optional, Tuple

from ..exceptions.exceptions import InvalidEventAttributeName, InvalidEventOperation
from .event import _NO_ATTRIBUTE_NAME_OPERATIONS, _UNSUBMITTABLE_ENTITY_TYPES, EventEntityType, EventOperation


class Topic:
    def __init__(
        self,
        entity_type: Optional[EventEntityType] = None,
        entity_id: Optional[str] = None,
        operation: Optional[EventOperation] = None,
        attribute_name: Optional[str] = None,
    ):

        self.entity_type = entity_type
        self.entity_id = entity_id
        self.operation = self.__preprocess_operation(operation, self.entity_type)
        self.attribute_name = self.__preprocess_attribute_name(attribute_name, self.operation)

    @classmethod
    def __preprocess_attribute_name(
        cls, attribute_name: Optional[str] = None, operation: Optional[EventOperation] = None
    ) -> Optional[str]:

        # if operation in _NO_ATTRIBUTE_NAME_OPERATIONS and attribute_name is not None:
        #     raise InvalidEventAttributeName
        return attribute_name

    @classmethod
    def __preprocess_operation(
        cls, operation: Optional[EventOperation] = None, entity_type: Optional[EventEntityType] = None
    ) -> Optional[EventOperation]:
        if (
            entity_type
            and operation
            and entity_type in _UNSUBMITTABLE_ENTITY_TYPES
            and operation == EventOperation.SUBMISSION
        ):
            raise InvalidEventOperation
        return operation

    def __hash__(self):
        return hash((self.entity_type, self.entity_id, self.operation, self.attribute_name))

    def __eq__(self, __value) -> bool:
        if (
            self.entity_type == __value.entity_type
            and self.entity_id == __value.entity_id
            and self.operation == __value.operation
            and self.attribute_name == __value.attribute_name
        ):
            return True
        return False



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import threading
from abc import abstractmethod
from typing import Any, Dict, List

from taipy.config._serializer._toml_serializer import _TomlSerializer
from taipy.config.config import Config
from taipy.logger._taipy_logger import _TaipyLogger

from ...config.job_config import JobConfig
from ...data._data_manager_factory import _DataManagerFactory
from ...data.data_node import DataNode
from ...exceptions.exceptions import DataNodeWritingError
from ...job._job_manager_factory import _JobManagerFactory
from ...job.job import Job
from ...job.job_id import JobId
from ...task.task import Task
from .._abstract_orchestrator import _AbstractOrchestrator


class _JobDispatcher(threading.Thread):
    """Manages job dispatching (instances of `Job^` class) on executors."""

    _STOP_FLAG = False
    _dispatched_processes: Dict = {}
    __logger = _TaipyLogger._get_logger()
    _nb_available_workers: int = 1

    def __init__(self, orchestrator: _AbstractOrchestrator):
        threading.Thread.__init__(self, name="Thread-Taipy-JobDispatcher")
        self.daemon = True
        self.orchestrator = orchestrator
        self.lock = self.orchestrator.lock  # type: ignore
        Config.block_update()

    def start(self):
        """Start the dispatcher"""
        threading.Thread.start(self)

    def is_running(self) -> bool:
        """Return True if the dispatcher is running"""
        return self.is_alive()

    def stop(self):
        """Stop the dispatcher"""
        self._STOP_FLAG = True

    def run(self):
        while not self._STOP_FLAG:
            try:
                if self._can_execute():
                    with self.lock:
                        job = self.orchestrator.jobs_to_run.get(block=True, timeout=0.1)
                    self._execute_job(job)
            except Exception:  # In case the last job of the queue has been removed.
                pass

    def _can_execute(self) -> bool:
        """Returns True if the dispatcher have resources to execute a new job."""
        return self._nb_available_workers > 0

    def _execute_job(self, job: Job):
        if job.force or self._needs_to_run(job.task):
            if job.force:
                self.__logger.info(f"job {job.id} is forced to be executed.")
            job.running()
            self._dispatch(job)
        else:
            job._unlock_edit_on_outputs()
            job.skipped()
            self.__logger.info(f"job {job.id} is skipped.")

    def _execute_jobs_synchronously(self):
        while not self.orchestrator.jobs_to_run.empty():
            with self.lock:
                try:
                    job = self.orchestrator.jobs_to_run.get()
                except Exception:  # In case the last job of the queue has been removed.
                    self.__logger.warning(f"{job.id} is no longer in the list of jobs to run.")
            self._execute_job(job)

    @staticmethod
    def _needs_to_run(task: Task) -> bool:
        """
        Returns True if the task has no output or if at least one input was modified since the latest run.

        Parameters:
             task (Task^): The task to run.
        Returns:
             True if the task needs to run. False otherwise.
        """
        if not task.skippable:
            return True
        data_manager = _DataManagerFactory._build_manager()
        if len(task.output) == 0:
            return True
        are_outputs_in_cache = all(data_manager._get(dn.id).is_up_to_date for dn in task.output.values())
        if not are_outputs_in_cache:
            return True
        if len(task.input) == 0:
            return False
        input_last_edit = max(data_manager._get(dn.id).last_edit_date for dn in task.input.values())
        output_last_edit = min(data_manager._get(dn.id).last_edit_date for dn in task.output.values())
        return input_last_edit > output_last_edit

    @abstractmethod
    def _dispatch(self, job: Job):
        """
        Dispatches the given `Job^` on an available worker for execution.

        Parameters:
            job (Job^): The job to submit on an executor with an available worker.
        """
        raise NotImplementedError

    @classmethod
    def _run_wrapped_function(cls, mode, config_as_string, job_id: JobId, task: Task):
        """
        Reads inputs, execute function, and write outputs.

        Parameters:
            mode: The job execution mode.
            config_as_string: The applied config passed as string to be reloaded iff the mode is `standalone`.
            job_id (JobId^): The id of the job.
            task (Task^): The task to be executed.
        Returns:
             True if the task needs to run. False otherwise.
        """
        try:
            if mode != JobConfig._DEVELOPMENT_MODE:
                Config._applied_config = _TomlSerializer()._deserialize(config_as_string)
                Config.block_update()
            inputs: List[DataNode] = list(task.input.values())
            outputs: List[DataNode] = list(task.output.values())

            fct = task.function

            results = fct(*cls.__read_inputs(inputs))
            return cls.__write_data(outputs, results, job_id)
        except Exception as e:
            return [e]

    @classmethod
    def __read_inputs(cls, inputs: List[DataNode]) -> List[Any]:
        data_manager = _DataManagerFactory._build_manager()
        return [data_manager._get(dn.id).read_or_raise() for dn in inputs]

    @classmethod
    def __write_data(cls, outputs: List[DataNode], results, job_id: JobId):
        data_manager = _DataManagerFactory._build_manager()
        try:
            if outputs:
                _results = cls.__extract_results(outputs, results)
                exceptions = []
                for res, dn in zip(_results, outputs):
                    try:
                        data_node = data_manager._get(dn.id)
                        data_node.write(res, job_id=job_id)
                        data_manager._set(data_node)
                    except Exception as e:
                        exceptions.append(DataNodeWritingError(f"Error writing in datanode id {dn.id}: {e}"))
                return exceptions
        except Exception as e:
            return [e]

    @classmethod
    def __extract_results(cls, outputs: List[DataNode], results: Any) -> List[Any]:
        _results: List[Any] = [results] if len(outputs) == 1 else results
        if len(_results) != len(outputs):
            raise DataNodeWritingError("Error: wrong number of result or task output")
        return _results

    @staticmethod
    def _update_job_status(job: Job, exceptions):
        job.update_status(exceptions)
        _JobManagerFactory._build_manager()._set(job)

    @classmethod
    def _set_dispatched_processes(cls, job_id, process):
        cls._dispatched_processes[job_id] = process

    @classmethod
    def _pop_dispatched_process(cls, job_id, default=None):
        return cls._dispatched_processes.pop(job_id, default)  # type: ignore



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import uuid
from abc import abstractmethod
from datetime import datetime, timedelta
from functools import reduce
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import modin.pandas as modin_pd
import numpy as np
import pandas as pd

from taipy.config.common._validate_id import _validate_id
from taipy.config.common.scope import Scope
from taipy.logger._taipy_logger import _TaipyLogger

from .._entity._entity import _Entity
from .._entity._labeled import _Labeled
from .._entity._properties import _Properties
from .._entity._reload import _reload, _self_reload, _self_setter
from .._version._version_manager_factory import _VersionManagerFactory
from ..common._warnings import _warn_deprecated
from ..exceptions.exceptions import NoData
from ..job.job_id import JobId
from ._filter import _FilterDataNode
from .data_node_id import DataNodeId, Edit
from .operator import JoinOperator, Operator


class DataNode(_Entity, _Labeled):
    """Reference to a dataset.

    A Data Node is an abstract class that holds metadata related to the dataset it refers to.
    In particular, a data node holds the name, the scope, the owner identifier, the last
    edit date, and some additional properties of the data.<br/>
    A Data Node also contains information and methods needed to access the dataset. This
    information depends on the type of storage, and it is held by subclasses (such as
    SQL Data Node, CSV Data Node, ...).

    !!! note
        It is recommended not to instantiate subclasses of `DataNode` directly.

    Attributes:
        config_id (str): Identifier of the data node configuration. It must be a valid Python
            identifier.
        scope (Scope^): The scope of this data node.
        id (str): The unique identifier of this data node.
        name (str): A user-readable name of this data node.
        owner_id (str): The identifier of the owner (pipeline_id, scenario_id, cycle_id) or
            None.
        parent_ids (Optional[Set[str]]): The set of identifiers of the parent tasks.
        last_edit_date (datetime): The date and time of the last modification.
        edits (List[Edit^]): The list of Edits (an alias for dict) containing medata about each
            edition of that node.
        version (str): The string indicates the application version of the data node to
            instantiate. If not provided, the current version is used.
        validity_period (Optional[timedelta]): The duration implemented as a timedelta since the last edit date for
            which the data node can be considered up-to-date. Once the validity period has passed, the data node is
            considered stale and relevant tasks will run even if they are skippable (see the
            [Task management page](../core/entities/task-mgt.md) for more details).
            If _validity_period_ is set to `None`, the data node is always up-to-date.
        edit_in_progress (bool): True if the data node is locked for modification. False
            otherwise.
        kwargs: A dictionary of additional properties.
    """

    _ID_PREFIX = "DATANODE"
    __ID_SEPARATOR = "_"
    __logger = _TaipyLogger._get_logger()
    _REQUIRED_PROPERTIES: List[str] = []
    _MANAGER_NAME = "data"
    __PATH_KEY = "path"

    def __init__(
        self,
        config_id,
        scope: Scope = Scope(Scope.SCENARIO),
        id: Optional[DataNodeId] = None,
        name: Optional[str] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        last_edit_date: Optional[datetime] = None,
        edits: Optional[List[Edit]] = None,
        version: Optional[str] = None,
        validity_period: Optional[timedelta] = None,
        edit_in_progress: bool = False,
        **kwargs,
    ):
        self.config_id = _validate_id(config_id)
        self.id = id or DataNodeId(self.__ID_SEPARATOR.join([self._ID_PREFIX, self.config_id, str(uuid.uuid4())]))
        self.owner_id = owner_id
        self._parent_ids = parent_ids or set()
        self._scope = scope
        self._last_edit_date = last_edit_date
        self._name = name
        self._edit_in_progress = edit_in_progress
        self._version = version or _VersionManagerFactory._build_manager()._get_latest_version()
        self._validity_period = validity_period

        # Track edits
        self._edits = edits or list()

        self._properties = _Properties(self, **kwargs)

    @property
    def parent_id(self):
        """Deprecated. Use `owner_id` instead."""
        _warn_deprecated("parent_id", suggest="owner_id")
        return self.owner_id

    @parent_id.setter
    def parent_id(self, val):
        """Deprecated. Use `owner_id` instead."""
        _warn_deprecated("parent_id", suggest="owner_id")
        self.owner_id = val

    def get_parents(self):
        """Get all parents of this data node."""
        from ... import core as tp

        return tp.get_parents(self)

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def parent_ids(self):
        """List of parent ids of this data node."""
        return self._parent_ids

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def edits(self):
        """Get all `Edit^`s of this data node."""
        return self._edits

    def get_last_edit(self) -> Optional[Edit]:
        """Get last `Edit^` of this data node.

        Returns:
            None if there has been no `Edit^` on this data node.
        """
        if self._edits:
            return self._edits[-1]
        return None

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def last_edit_date(self):
        last_modified_datetime = self.__get_last_modified_datetime()
        if last_modified_datetime and last_modified_datetime > self._last_edit_date:
            return last_modified_datetime
        else:
            return self._last_edit_date

    @last_edit_date.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def last_edit_date(self, val):
        self._last_edit_date = val

    @property
    def last_edition_date(self):
        """Deprecated. Use `last_edit_date` instead."""
        _warn_deprecated("last_edition_date", suggest="last_edit_date")
        return self.last_edit_date

    @last_edition_date.setter
    def last_edition_date(self, val):
        _warn_deprecated("last_edition_date", suggest="last_edit_date")
        self.last_edit_date = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def scope(self):
        if self._scope == Scope.PIPELINE:
            _warn_deprecated(f"`{Scope.PIPELINE}`", suggest="other `Scope` value")
        return self._scope

    @scope.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def scope(self, val):
        self._scope = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def validity_period(self) -> Optional[timedelta]:
        return self._validity_period if self._validity_period else None

    @validity_period.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def validity_period(self, val):
        self._validity_period = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def expiration_date(self) -> datetime:
        """Datetime instant of the expiration date of this data node."""
        last_edit_date = self.last_edit_date
        validity_period = self._validity_period

        if not last_edit_date:
            raise NoData(f"Data node {self.id} from config {self.config_id} has not been written yet.")

        return last_edit_date + validity_period if validity_period else last_edit_date

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def name(self):
        return self._name

    @name.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def name(self, val):
        self._name = val

    @property
    def version(self):
        return self._version

    @property
    def cacheable(self):
        """Deprecated. Use `skippable` attribute of a `Task^` instead."""
        _warn_deprecated("cacheable", suggest="the skippable feature")
        return self.properties.get("cacheable", False)

    @cacheable.setter
    def cacheable(self, val):
        _warn_deprecated("cacheable", suggest="the skippable feature")

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def edit_in_progress(self):
        return self._edit_in_progress

    @edit_in_progress.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def edit_in_progress(self, val):
        self._edit_in_progress = val

    @property
    def edition_in_progress(self):
        """Deprecated. Use `edit_in_progress` instead."""
        _warn_deprecated("edition_in_progress", suggest="edit_in_progress")
        return self.edit_in_progress

    @edition_in_progress.setter
    def edition_in_progress(self, val):
        """Deprecated. Use `edit_in_progress` instead."""
        _warn_deprecated("edition_in_progress", suggest="edit_in_progress")
        self.edit_in_progress = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def job_ids(self):
        """List of the jobs having edited this data node."""
        return [edit.get("job_id") for edit in self.edits if edit.get("job_id")]

    @property
    def properties(self):
        """Dictionary of custom properties."""
        r = _reload(self._MANAGER_NAME, self)
        self._properties = r._properties
        return self._properties

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.id)

    def __getstate__(self):
        return vars(self)

    def __setstate__(self, state):
        vars(self).update(state)

    def __getattr__(self, attribute_name):
        protected_attribute_name = _validate_id(attribute_name)
        if protected_attribute_name in self._properties:
            return self._properties[protected_attribute_name]
        raise AttributeError(f"{attribute_name} is not an attribute of data node {self.id}")

    def __get_last_modified_datetime(self) -> Optional[datetime]:
        path = self._properties.get(self.__PATH_KEY, None)
        if path and os.path.isfile(path):
            return datetime.fromtimestamp(os.path.getmtime(path))

        last_modified_datetime = None
        if path and os.path.isdir(path):
            for filename in os.listdir(path):
                filepath = os.path.join(path, filename)
                if os.path.isfile(filepath):
                    file_mtime = datetime.fromtimestamp(os.path.getmtime(filepath))

                    if last_modified_datetime is None or file_mtime > last_modified_datetime:
                        last_modified_datetime = file_mtime

        return last_modified_datetime

    @classmethod
    @abstractmethod
    def storage_type(cls) -> str:
        return NotImplementedError  # type: ignore

    def read_or_raise(self) -> Any:
        """Read the data referenced by this data node.

        Returns:
            The data referenced by this data node.

        Raises:
            NoData^: If the data has not been written yet.
        """
        if not self.last_edit_date:
            raise NoData(f"Data node {self.id} from config {self.config_id} has not been written yet.")
        return self._read()

    def read(self) -> Any:
        """Read the data referenced by this data node.

        Returns:
            The data referenced by this data node. None if the data has not been written yet.
        """
        try:
            return self.read_or_raise()
        except NoData:
            self.__logger.warning(
                f"Data node {self.id} from config {self.config_id} is being read but has never been " f"written."
            )
            return None

    def write(self, data, job_id: Optional[JobId] = None, **kwargs: Dict[str, Any]):
        """Write some data to this data node.

        Parameters:
            data (Any): The data to write to this data node.
            job_id (JobId^): An optional identifier of the writer.
            **kwargs (dict[str, any]): Extra information to attach to the edit document
                corresponding to this write.
        """
        from ._data_manager_factory import _DataManagerFactory

        self._write(data)
        self._track_edit(job_id=job_id, **kwargs)
        self.unlock_edit()
        _DataManagerFactory._build_manager()._set(self)

    def _track_edit(self, **options):
        """Add Edit tracking information to this data node."""
        edit = {}
        for k, v in options.items():
            if v is not None:
                edit[k] = v
        if "timestamp" not in edit:
            edit["timestamp"] = datetime.now()
        self.last_edit_date = edit.get("timestamp")
        self._edits.append(edit)

    def lock_edit(self):
        """Lock the data node modification.

        Note:
            The data node can be unlocked with the method `(DataNode.)unlock_edit()^`.
        """
        self.edit_in_progress = True

    def lock_edition(self):
        """Deprecated. Use `lock_edit` instead."""
        _warn_deprecated("lock_edition", suggest="lock_edit")
        self.lock_edit()

    def unlock_edit(self, at: Optional[datetime] = None, job_id: Optional[JobId] = None):
        """Unlocks the data node modification.

        Parameters:
            at (Optional[datetime]): Deprecated.
            job_id (Optional[JobId^]): Deprecated.

        Note:
            The data node can be locked with the method `(DataNode.)lock_edit()^`.
        """
        self.edit_in_progress = False  # type: ignore

    def unlock_edition(self, at: Optional[datetime] = None, job_id: Optional[JobId] = None):
        """Deprecated. Use `(DataNode.)unlock_edit()^` instead."""
        _warn_deprecated("unlock_edition", suggest="unlock_edit")
        self.unlock_edit()

    def filter(self, operators: Union[List, Tuple], join_operator=JoinOperator.AND):
        """Read and filter the data referenced by this data node.

        The data is filtered by the provided list of 3-tuples (key, value, `Operator^`).

        If multiple filter operators are provided, filtered data will be joined based on the
        join operator (_AND_ or _OR_).

        Parameters:
            operators (Union[List[Tuple], Tuple]): TODO
            join_operator (JoinOperator^): The operator used to join the multiple filter
                3-tuples.
        """
        data = self._read()
        if len(operators) == 0:
            return data
        if not ((type(operators[0]) == list) or (type(operators[0]) == tuple)):
            if isinstance(data, (pd.DataFrame, modin_pd.DataFrame)):
                return DataNode.__filter_dataframe_per_key_value(data, operators[0], operators[1], operators[2])
            if isinstance(data, List):
                return DataNode.__filter_list_per_key_value(data, operators[0], operators[1], operators[2])
        else:
            if isinstance(data, (pd.DataFrame, modin_pd.DataFrame)):
                return DataNode.__filter_dataframe(data, operators, join_operator=join_operator)
            if isinstance(data, List):
                return DataNode.__filter_list(data, operators, join_operator=join_operator)
        return NotImplementedError

    @staticmethod
    def __filter_dataframe(
        df_data: Union[pd.DataFrame, modin_pd.DataFrame], operators: Union[List, Tuple], join_operator=JoinOperator.AND
    ):
        filtered_df_data = []
        if join_operator == JoinOperator.AND:
            how = "inner"
        elif join_operator == JoinOperator.OR:
            how = "outer"
        else:
            return NotImplementedError
        for key, value, operator in operators:
            filtered_df_data.append(DataNode.__filter_dataframe_per_key_value(df_data, key, value, operator))
        return DataNode.__dataframe_merge(filtered_df_data, how) if filtered_df_data else pd.DataFrame()

    @staticmethod
    def __filter_dataframe_per_key_value(
        df_data: Union[pd.DataFrame, modin_pd.DataFrame], key: str, value, operator: Operator
    ):
        df_by_col = df_data[key]
        if operator == Operator.EQUAL:
            df_by_col = df_by_col == value
        if operator == Operator.NOT_EQUAL:
            df_by_col = df_by_col != value
        if operator == Operator.LESS_THAN:
            df_by_col = df_by_col < value
        if operator == Operator.LESS_OR_EQUAL:
            df_by_col = df_by_col <= value
        if operator == Operator.GREATER_THAN:
            df_by_col = df_by_col > value
        if operator == Operator.GREATER_OR_EQUAL:
            df_by_col = df_by_col >= value
        return df_data[df_by_col]

    @staticmethod
    def __dataframe_merge(df_list: List, how="inner"):
        return reduce(lambda df1, df2: pd.merge(df1, df2, how=how), df_list)

    @staticmethod
    def __filter_list(list_data: List, operators: Union[List, Tuple], join_operator=JoinOperator.AND):
        filtered_list_data = []
        for key, value, operator in operators:
            filtered_list_data.append(DataNode.__filter_list_per_key_value(list_data, key, value, operator))
        if len(filtered_list_data) == 0:
            return filtered_list_data
        if join_operator == JoinOperator.AND:
            return DataNode.__list_intersect(filtered_list_data)
        elif join_operator == JoinOperator.OR:
            return list(set(np.concatenate(filtered_list_data)))
        else:
            return NotImplementedError

    @staticmethod
    def __filter_list_per_key_value(list_data: List, key: str, value, operator: Operator):
        filtered_list = []
        for row in list_data:
            row_value = getattr(row, key)
            if operator == Operator.EQUAL and row_value == value:
                filtered_list.append(row)
            if operator == Operator.NOT_EQUAL and row_value != value:
                filtered_list.append(row)
            if operator == Operator.LESS_THAN and row_value < value:
                filtered_list.append(row)
            if operator == Operator.LESS_OR_EQUAL and row_value <= value:
                filtered_list.append(row)
            if operator == Operator.GREATER_THAN and row_value > value:
                filtered_list.append(row)
            if operator == Operator.GREATER_OR_EQUAL and row_value >= value:
                filtered_list.append(row)
        return filtered_list

    @staticmethod
    def __list_intersect(list_data):
        return list(set(list_data.pop()).intersection(*map(set, list_data)))

    @abstractmethod
    def _read(self):
        raise NotImplementedError

    @abstractmethod
    def _write(self, data):
        raise NotImplementedError

    def __getitem__(self, items):
        return _FilterDataNode(self.id, self._read())[items]

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def is_ready_for_reading(self) -> bool:
        """Indicate if this data node is ready for reading.

        Returns:
            False if the data is locked for modification or if the data has never been written.
                True otherwise.
        """
        if self._edit_in_progress:
            return False
        if not self._last_edit_date:
            # Never been written so it is not up-to-date
            return False
        return True

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def is_up_to_date(self) -> bool:
        """Indicate if this data node is up-to-date.

        Returns:
            False if the data ever been written or the expiration date has passed.<br/>
            True otherwise.
        """
        if not self._last_edit_date:
            # Never been written so it is not up-to-date
            return False
        if not self._validity_period:
            # No validity period and has already been written, so it is up-to-date
            return True
        if datetime.now() > self.expiration_date:
            # expiration_date has been passed
            return False
        return True

    @staticmethod
    def _class_map():
        def all_subclasses(cls):
            subclasses = set(cls.__subclasses__())
            for s in cls.__subclasses__():
                subclasses.update(all_subclasses(s))
            return subclasses

        return {c.storage_type(): c for c in all_subclasses(DataNode) if c.storage_type() is not None}

    def get_label(self) -> str:
        """Returns the data node simple label prefixed by its owner label.

        Returns:
            The label of the data node as a string.
        """
        return self._get_label()

    def get_simple_label(self) -> str:
        """Returns the data node simple label.

        Returns:
            The simple label of the data node as a string.
        """
        return self._get_simple_label()



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from .csv import CSVDataNode
from .data_node import DataNode
from .excel import ExcelDataNode
from .generic import GenericDataNode
from .in_memory import InMemoryDataNode
from .json import JSONDataNode
from .mongo import MongoCollectionDataNode
from .operator import JoinOperator, Operator
from .parquet import ParquetDataNode
from .pickle import PickleDataNode
from .sql import SQLDataNode
from .sql_table import SQLTableDataNode



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from copy import copy
from typing import Any, Dict, Optional

from taipy.config import Config, UniqueSection
from taipy.config._config import _Config
from taipy.config.common._template_handler import _TemplateHandler as _tpl


class CoreSection(UniqueSection):
    name = "CORE"
    _MODE_KEY = "mode"
    _DEVELOPMENT_MODE = "development"
    _EXPERIMENT_MODE = "experiment"
    _PRODUCTION_MODE = "production"
    _DEFAULT_MODE = _DEVELOPMENT_MODE
    _VERSION_NUMBER_KEY = "version_number"
    _DEFAULT_VERSION_NUMBER = ""
    _TAIPY_FORCE_KEY = "force"
    _DEFAULT_TAIPY_FORCE = False
    _CLEAN_ENTITIES_KEY = "clean_entities"
    _DEFAULT_CLEAN_ENTITIES = False

    def __init__(
        self,
        mode: Optional[str] = None,
        version_number: Optional[str] = None,
        force: Optional[bool] = None,
        clean_entities: Optional[bool] = None,
        **properties,
    ):
        self.mode = mode or self._DEFAULT_MODE
        self.version_number = version_number or self._DEFAULT_VERSION_NUMBER
        self.force = force or self._DEFAULT_TAIPY_FORCE
        self.clean_entities = clean_entities or self._DEFAULT_CLEAN_ENTITIES
        super().__init__(**properties)

    def __copy__(self):
        return CoreSection(self.mode, self.version_number, self.force, self.clean_entities, **copy(self._properties))

    @classmethod
    def default_config(cls):
        return CoreSection(
            cls._DEFAULT_MODE, cls._DEFAULT_VERSION_NUMBER, cls._DEFAULT_TAIPY_FORCE, cls._DEFAULT_CLEAN_ENTITIES
        )

    def _clean(self):
        self.mode = self._DEFAULT_MODE
        self.version_number = self._DEFAULT_VERSION_NUMBER
        self.force = self._DEFAULT_TAIPY_FORCE
        self.clean_entities = self._DEFAULT_CLEAN_ENTITIES
        self._properties.clear()

    def _to_dict(self):
        as_dict = {}
        if self.mode is not None:
            as_dict[self._MODE_KEY] = self.mode
        if self.version_number is not None:
            as_dict[self._VERSION_NUMBER_KEY] = self.version_number
        if self.force is not None:
            as_dict[self._TAIPY_FORCE_KEY] = self.force
        if self.clean_entities is not None:
            as_dict[self._CLEAN_ENTITIES_KEY] = self.clean_entities
        as_dict.update(self._properties)
        return as_dict

    @classmethod
    def _from_dict(cls, as_dict: Dict[str, Any], id=None, config: Optional[_Config] = None):
        mode = as_dict.pop(cls._MODE_KEY, None)
        version_nb = as_dict.pop(cls._VERSION_NUMBER_KEY, None)
        force = as_dict.pop(cls._TAIPY_FORCE_KEY, None)
        clean_entities = as_dict.pop(cls._CLEAN_ENTITIES_KEY, None)
        return CoreSection(mode, version_nb, force, clean_entities, **as_dict)

    def _update(self, as_dict: Dict[str, Any]):
        mode = _tpl._replace_templates(as_dict.pop(self._MODE_KEY, self.mode))
        if self.mode != mode:
            self.mode = mode
        version_number = _tpl._replace_templates(as_dict.pop(self._VERSION_NUMBER_KEY, self.version_number))
        if self.version_number != version_number:
            self.version_number = version_number
        force = _tpl._replace_templates(as_dict.pop(self._TAIPY_FORCE_KEY, self.force))
        if self.force != force:
            self.force = force
        clean_entities = _tpl._replace_templates(as_dict.pop(self._CLEAN_ENTITIES_KEY, self.clean_entities))
        if self.clean_entities != clean_entities:
            self.clean_entities = clean_entities
        self._properties.update(as_dict)

    @staticmethod
    def _configure(
        mode: Optional[str] = None,
        version_number: Optional[str] = None,
        force: Optional[bool] = None,
        clean_entities: Optional[bool] = None,
        **properties,
    ) -> "CoreSection":
        """Configure the Core service.

        Parameters:
            mode (Optional[str], optional): Indicates the mode of the version management system.
                Possible values are *"development"*, *"experiment"*, or *"production"*.
            version_number (Optional[str], optional): The string identifier of the version.
                 In development mode, the version number is ignored.
            force (Optional[bool], optional): If True, Taipy will override a version even if the configuration
                has changed and run the application.
            clean_entities (Optional[bool], optional): If True, running a Taipy Core service will clean all current
                version entities before running the application.
            **properties (Dict[str, Any]): A keyworded variable length list of additional arguments configure the
                behavior of the `Core^` service.
        Returns:
            The Core configuration.
        """
        section = CoreSection(
            mode=mode, version_number=version_number, force=force, clean_entities=clean_entities, **properties
        )
        Config._register(section)
        return Config.unique_sections[CoreSection.name]



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy.config._config import _Config
from taipy.config.checker._checkers._config_checker import _ConfigChecker
from taipy.config.checker.issue_collector import IssueCollector

from ..pipeline_config import PipelineConfig
from ..task_config import TaskConfig


class _PipelineConfigChecker(_ConfigChecker):
    def __init__(self, config: _Config, collector: IssueCollector):
        super().__init__(config, collector)

    def _check(self) -> IssueCollector:
        pipeline_configs = self._config._sections[PipelineConfig.name]
        for pipeline_config_id, pipeline_config in pipeline_configs.items():
            if pipeline_config_id != _Config.DEFAULT_KEY:
                self._check_if_entity_property_key_used_is_predefined(pipeline_config)
                self._check_existing_config_id(pipeline_config)
                self._check_tasks(pipeline_config_id, pipeline_config)
        return self._collector

    def _check_tasks(self, pipeline_config_id: str, pipeline_config: PipelineConfig):
        self._check_children(
            PipelineConfig, pipeline_config_id, pipeline_config._TASK_KEY, pipeline_config.tasks, TaskConfig
        )



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from queue import SimpleQueue
from typing import Optional
from uuid import uuid4

from .event import EventEntityType, EventOperation
from .registration_id import RegistrationId
from .topic import Topic


class Registration:

    _ID_PREFIX = "REGISTRATION"
    __SEPARATOR = "_"

    def __init__(
        self,
        entity_type: Optional[EventEntityType] = None,
        entity_id: Optional[str] = None,
        operation: Optional[EventOperation] = None,
        attribute_name: Optional[str] = None,
    ):

        self.registration_id: str = self._new_id()
        self.topic: Topic = Topic(entity_type, entity_id, operation, attribute_name)
        self.queue: SimpleQueue = SimpleQueue()

    @staticmethod
    def _new_id() -> RegistrationId:
        """Generate a unique scenario identifier."""
        return RegistrationId(Registration.__SEPARATOR.join([Registration._ID_PREFIX, str(uuid4())]))

    def __hash__(self) -> int:
        return hash(self.registration_id)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy._cli._base_cli import _CLI

from .config import CoreSection


class _CoreCLI:
    """Command-line interface for Taipy Core application."""

    @classmethod
    def create_parser(cls):
        core_parser = _CLI._add_groupparser("Taipy Core", "Optional arguments for Taipy Core service")

        mode_group = core_parser.add_mutually_exclusive_group()
        mode_group.add_argument(
            "--development",
            action="store_true",
            help="""
                When execute Taipy application in `development` mode, all entities from the previous development version
                will be deleted before running new Taipy application.
            """,
        )
        mode_group.add_argument(
            "--experiment",
            nargs="?",
            const="",
            metavar="VERSION",
            help="""
                When execute Taipy application in `experiment` mode, the current Taipy application is saved to a new
                version. If version name already exists, check for compatibility with current Python Config and run the
                application. Without being specified, the version number will be a random string.
            """,
        )
        mode_group.add_argument(
            "--production",
            nargs="?",
            const="",
            metavar="VERSION",
            help="""
                When execute in `production` mode, the current version is used in production. All production versions
                should have the same configuration and share all entities. Without being specified, the latest version
                is used.
            """,
        )

        force_group = core_parser.add_mutually_exclusive_group()
        force_group.add_argument(
            "--taipy-force",
            action="store_true",
            help="Force override the configuration of the version if existed and run the application."
            " Default to False.",
        )
        force_group.add_argument(
            "--no-taipy-force",
            action="store_true",
            help="Stop the application if any Config conflict exists.",
        )

        clean_entities_group = core_parser.add_mutually_exclusive_group()
        clean_entities_group.add_argument(
            "--clean-entities",
            action="store_true",
            help="Clean all current version entities before running the application. Default to False.",
        )
        clean_entities_group.add_argument(
            "--no-clean-entities",
            action="store_true",
            help="Keep all entities of the current experiment version.",
        )

    @classmethod
    def parse_arguments(cls):
        args = _CLI._parse()
        as_dict = {}
        if args.development:
            as_dict[CoreSection._MODE_KEY] = CoreSection._DEVELOPMENT_MODE
        elif args.experiment is not None:
            as_dict[CoreSection._MODE_KEY] = CoreSection._EXPERIMENT_MODE
            as_dict[CoreSection._VERSION_NUMBER_KEY] = args.experiment
        elif args.production is not None:
            as_dict[CoreSection._MODE_KEY] = CoreSection._PRODUCTION_MODE
            as_dict[CoreSection._VERSION_NUMBER_KEY] = args.production

        if args.taipy_force:
            as_dict[CoreSection._TAIPY_FORCE_KEY] = True
        elif args.no_taipy_force:
            as_dict[CoreSection._TAIPY_FORCE_KEY] = False

        if args.clean_entities:
            as_dict[CoreSection._CLEAN_ENTITIES_KEY] = True
        elif args.no_clean_entities:
            as_dict[CoreSection._CLEAN_ENTITIES_KEY] = False
        return as_dict



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime
from typing import Any

from taipy.config import Config
from taipy.config._config import _Config

from .._entity._entity import _Entity


class _Version(_Entity):
    def __init__(self, id: str, config: Any) -> None:
        self.id: str = id
        self.config: _Config = config
        self.creation_date: datetime = datetime.now()

    def __eq__(self, other):
        return self.id == other.id and self.__is_config_eq(other)

    def __is_config_eq(self, other):
        return Config._serializer._str(self.config) == Config._serializer._str(other.config)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy.config.common._validate_id import _validate_id


class DefaultCustomDocument:
    """The default class for \"custom_document\" property to configure a `MongoCollectionDataNode^`.

    Attributes:
        **kwargs: Attributes of the DefaultCustomDocument object.

    Example:
        - `document = DefaultCustomDocument(name="example", age=30})`
        will return a DefaultCustomDocument object so that `document.name` returns `"example"`,
        and `document.age` returns `30`.
        - `document = DefaultCustomDocument(date="12/24/2018", temperature=20})`
        will return a DefaultCustomDocument object so that `document.date` returns `"12/24/2018"`,
        and `document.temperature` returns `20`.
    """

    def __init__(self, **kwargs):
        for attribute_name, value in kwargs.items():
            setattr(self, _validate_id(attribute_name), value)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import dataclasses
import json
import os
from datetime import date, datetime, timedelta
from enum import Enum
from os.path import isfile
from pydoc import locate
from typing import Any, Dict, List, Optional, Set

from taipy.config.common.scope import Scope

from .._backup._backup import _replace_in_backup_file
from .._entity._reload import _self_reload
from .._version._version_manager_factory import _VersionManagerFactory
from .abstract_file import _AbstractFileDataNode
from .data_node import DataNode
from .data_node_id import DataNodeId, Edit


class JSONDataNode(DataNode, _AbstractFileDataNode):
    """Data Node stored as a JSON file.

    Attributes:
        config_id (str): Identifier of the data node configuration. This string must be a valid
            Python identifier.
        scope (Scope^): The scope of this data node.
        id (str): The unique identifier of this data node.
        name (str): A user-readable name of this data node.
        owner_id (str): The identifier of the owner (pipeline_id, scenario_id, cycle_id) or `None`.
        parent_ids (Optional[Set[str]]): The identifiers of the parent tasks or `None`.
        last_edit_date (datetime): The date and time of the last modification.
        edits (List[Edit^]): The ordered list of edits for that job.
        version (str): The string indicates the application version of the data node to instantiate. If not provided,
            the current version is used.
        validity_period (Optional[timedelta]): The duration implemented as a timedelta since the last edit date for
            which the data node can be considered up-to-date. Once the validity period has passed, the data node is
            considered stale and relevant tasks will run even if they are skippable (see the
            [Task management page](../core/entities/task-mgt.md) for more details).
            If _validity_period_ is set to `None`, the data node is always up-to-date.
        edit_in_progress (bool): True if a task computing the data node has been submitted
            and not completed yet. False otherwise.
        path (str): The path to the JSON file.
        encoder (json.JSONEncoder): The JSON encoder that is used to write into the JSON file.
        decoder (json.JSONDecoder): The JSON decoder that is used to read from the JSON file.
        properties (dict[str, Any]): A dictionary of additional properties. The _properties_
            must have a _"default_path"_ or _"path"_ entry with the path of the JSON file.
    """

    __STORAGE_TYPE = "json"
    __DEFAULT_DATA_KEY = "default_data"
    __DEFAULT_PATH_KEY = "default_path"
    __PATH_KEY = "path"
    _ENCODER_KEY = "encoder"
    _DECODER_KEY = "decoder"
    _REQUIRED_PROPERTIES: List[str] = []

    def __init__(
        self,
        config_id: str,
        scope: Scope,
        id: Optional[DataNodeId] = None,
        name: Optional[str] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        last_edit_date: Optional[datetime] = None,
        edits: List[Edit] = None,
        version: str = None,
        validity_period: Optional[timedelta] = None,
        edit_in_progress: bool = False,
        properties: Dict = None,
    ):
        if properties is None:
            properties = {}

        default_value = properties.pop(self.__DEFAULT_DATA_KEY, None)

        super().__init__(
            config_id,
            scope,
            id,
            name,
            owner_id,
            parent_ids,
            last_edit_date,
            edits,
            version or _VersionManagerFactory._build_manager()._get_latest_version(),
            validity_period,
            edit_in_progress,
            **properties,
        )
        self._path = properties.get(self.__PATH_KEY, properties.get(self.__DEFAULT_PATH_KEY))
        if not self._path:
            self._path = self._build_path(self.storage_type())
        properties[self.__PATH_KEY] = self._path

        self._decoder = self._properties.get(self._DECODER_KEY, _DefaultJSONDecoder)
        self._encoder = self._properties.get(self._ENCODER_KEY, _DefaultJSONEncoder)

        if default_value is not None and not os.path.exists(self._path):
            self.write(default_value)

        if not self._last_edit_date and isfile(self._path):  # type: ignore
            self.last_edit_date = datetime.now()  # type: ignore

    @classmethod
    def storage_type(cls) -> str:
        return cls.__STORAGE_TYPE

    @property  # type: ignore
    @_self_reload(DataNode._MANAGER_NAME)
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        tmp_old_path = self._path
        self._path = value
        self.properties[self.__PATH_KEY] = value
        _replace_in_backup_file(old_file_path=tmp_old_path, new_file_path=self._path)

    @property  # type: ignore
    @_self_reload(DataNode._MANAGER_NAME)
    def encoder(self):
        return self._encoder

    @encoder.setter
    def encoder(self, encoder: json.JSONEncoder):
        self.properties[self._ENCODER_KEY] = encoder

    @property  # type: ignore
    @_self_reload(DataNode._MANAGER_NAME)
    def decoder(self):
        return self._decoder

    @decoder.setter
    def decoder(self, decoder: json.JSONDecoder):
        self.properties[self._DECODER_KEY] = decoder

    def _read(self):
        with open(self._path, "r") as f:
            return json.load(f, cls=self._decoder)

    def _write(self, data: Any):
        with open(self._path, "w") as f:  # type: ignore
            json.dump(data, f, indent=4, cls=self._encoder)


class _DefaultJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Enum):
            return {
                "__type__": f"Enum-{o.__class__.__module__}-{o.__class__.__qualname__}-{o.name}",
                "__value__": o.value,
            }

        if isinstance(o, (datetime, date)):
            return {"__type__": "Datetime", "__value__": o.isoformat()}

        if dataclasses.is_dataclass(o):
            return {
                "__type__": f"dataclass-{o.__class__.__module__}-{o.__class__.__qualname__}",
                "__value__": dataclasses.asdict(o),
            }

        return super().default(o)


class _DefaultJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, source):
        if _type := source.get("__type__"):
            if _type.startswith("Enum"):
                _, module, classname, name = _type.split("-")
                _enum_class = locate(f"{module}.{classname}")
                return _enum_class[name]

            if _type == "Datetime":
                return datetime.fromisoformat(source.get("__value__"))

            if _type.startswith("dataclass"):
                _, module, classname = _type.split("-")
                _data_class = locate(f"{module}.{classname}")
                return _data_class(**source.get("__value__"))

        return source



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from abc import abstractmethod
from typing import Callable, Iterable, List, Optional, Union

from ..job.job import Job
from ..task.task import Task


class _AbstractOrchestrator:
    """Creates, enqueues, and orchestrates jobs as instances of `Job^` class."""

    @classmethod
    @abstractmethod
    def initialize(cls):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def submit(
        cls,
        pipeline,
        callbacks: Optional[Iterable[Callable]],
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
    ) -> List[Job]:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def submit_task(
        cls,
        task: Task,
        submit_id: Optional[str] = None,
        callbacks: Optional[Iterable[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
    ) -> Job:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def cancel_job(cls, job):
        raise NotImplementedError



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import sys


def _vt_codes_enabled_in_windows_registry():
    """
    Check the Windows Registry to see if VT code handling has been enabled
    by default, see https://superuser.com/a/1300251/447564.
    """
    try:
        # winreg is only available on Windows.
        import winreg
    except ImportError:
        return False
    else:
        try:
            reg_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, "Console")
            reg_key_value, _ = winreg.QueryValueEx(reg_key, "VirtualTerminalLevel")
        except FileNotFoundError:
            return False
        else:
            return reg_key_value == 1


def _is_color_supported():
    """
    Return True if the running system's terminal supports color,
    and False otherwise.
    """

    is_a_tty = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()

    return is_a_tty and (
        sys.platform != "win32"
        or "ANSICON" in os.environ
        or "WT_SESSION" in os.environ  # Windows Terminal supports VT codes.
        or os.environ.get("TERM_PROGRAM") == "vscode"  # VSCode's built-in terminal supports colors.
        or _vt_codes_enabled_in_windows_registry()
    )


class _Bcolors:
    PURPLE = "\033[95m" if _is_color_supported() else ""
    BLUE = "\033[94m" if _is_color_supported() else ""
    CYAN = "\033[96m" if _is_color_supported() else ""
    GREEN = "\033[92m" if _is_color_supported() else ""
    BOLD = "\033[1m" if _is_color_supported() else ""
    UNDERLINE = "\033[4m" if _is_color_supported() else ""
    END = "\033[0m" if _is_color_supported() else ""



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy.config._serializer._toml_serializer import _TomlSerializer
from taipy.config.config import Config

from ...job.job import Job
from .._abstract_orchestrator import _AbstractOrchestrator
from ._job_dispatcher import _JobDispatcher


class _DevelopmentJobDispatcher(_JobDispatcher):
    """Manages job dispatching (instances of `Job^` class) in a synchronous way."""

    def __init__(self, orchestrator: _AbstractOrchestrator):
        super().__init__(orchestrator)

    def start(self):
        raise NotImplementedError

    def is_running(self) -> bool:
        return True

    def stop(self):
        raise NotImplementedError

    def run(self):
        raise NotImplementedError

    def _dispatch(self, job: Job):
        """Dispatches the given `Job^` on an available worker for execution.

        Parameters:
            job (Job^): The job to submit on an executor with an available worker.
        """
        config_as_string = _TomlSerializer()._serialize(Config._applied_config)

        rs = self._run_wrapped_function(Config.job_config.mode, config_as_string, job.id, job.task)
        self._update_job_status(job, rs)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from enum import Enum


class Operator(Enum):
    """Enumeration of operators for Data Node filtering.

    The possible values are:

    - `EQUAL`
    - `NOT_EQUAL`
    - `LESS_THAN`
    - `LESS_OR_EQUAL`
    - `GREATER_THAN`
    - `GREATER_OR_EQUAL`
    """

    EQUAL = 1
    NOT_EQUAL = 2
    LESS_THAN = 3
    LESS_OR_EQUAL = 4
    GREATER_THAN = 5
    GREATER_OR_EQUAL = 6


class JoinOperator(Enum):
    """
    Enumeration of join operators for Data Node filtering. The possible values are `AND` and `OR`.
    """

    AND = 1
    OR = 2



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime
from typing import Optional

from ..common._repr_enum import _ReprEnum
from ..exceptions.exceptions import InvalidEventAttributeName, InvalidEventOperation


class EventOperation(_ReprEnum):
    CREATION = 1
    UPDATE = 2
    DELETION = 3
    SUBMISSION = 4


class EventEntityType(_ReprEnum):
    CYCLE = 1
    SCENARIO = 2
    PIPELINE = 3
    TASK = 4
    DATA_NODE = 5
    JOB = 6


_NO_ATTRIBUTE_NAME_OPERATIONS = set([EventOperation.CREATION, EventOperation.DELETION, EventOperation.SUBMISSION])
_UNSUBMITTABLE_ENTITY_TYPES = (EventEntityType.CYCLE, EventEntityType.DATA_NODE, EventEntityType.JOB)
_ENTITY_TO_EVENT_ENTITY_TYPE = {
    "scenario": EventEntityType.SCENARIO,
    "pipeline": EventEntityType.PIPELINE,
    "task": EventEntityType.TASK,
    "data": EventEntityType.DATA_NODE,
    "job": EventEntityType.JOB,
    "cycle": EventEntityType.CYCLE,
}


class Event:
    def __init__(
        self,
        entity_type: EventEntityType,
        entity_id: Optional[str],
        operation: EventOperation,
        attribute_name: Optional[str] = None,
    ):
        self.creation_date = datetime.now()
        self.entity_id = entity_id
        self.entity_type = entity_type
        self.operation = self.__preprocess_operation(operation, entity_type)
        self.attribute_name = self.__preprocess_attribute_name(attribute_name, operation)

    @classmethod
    def __preprocess_attribute_name(cls, attribute_name: Optional[str], operation: EventOperation) -> Optional[str]:
        if operation in _NO_ATTRIBUTE_NAME_OPERATIONS and attribute_name is not None:
            raise InvalidEventAttributeName
        return attribute_name

    @classmethod
    def __preprocess_operation(cls, operation: EventOperation, entity_type: EventEntityType) -> EventOperation:
        if entity_type in _UNSUBMITTABLE_ENTITY_TYPES and operation == EventOperation.SUBMISSION:
            raise InvalidEventOperation
        return operation



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from functools import partial
from typing import Any, Callable, List, Optional, Union

from taipy import Config
from taipy.config.common.scope import Scope

from .._entity._entity_ids import _EntityIds
from .._manager._manager import _Manager
from .._version._version_manager_factory import _VersionManagerFactory
from ..common.warn_if_inputs_not_ready import _warn_if_inputs_not_ready
from ..config.pipeline_config import PipelineConfig
from ..cycle.cycle_id import CycleId
from ..exceptions.exceptions import NonExistingPipeline
from ..job._job_manager_factory import _JobManagerFactory
from ..job.job import Job
from ..notification import EventEntityType, EventOperation, _publish_event
from ..scenario.scenario_id import ScenarioId
from ..task._task_manager_factory import _TaskManagerFactory
from ..task.task import Task
from ._pipeline_repository_factory import _PipelineRepositoryFactory
from .pipeline import Pipeline
from .pipeline_id import PipelineId


class _PipelineManager(_Manager[Pipeline]):
    _repository = _PipelineRepositoryFactory._build_repository()
    _ENTITY_NAME = Pipeline.__name__
    _EVENT_ENTITY_TYPE = EventEntityType.PIPELINE

    @classmethod
    def _subscribe(
        cls,
        callback: Callable[[Pipeline, Job], None],
        params: Optional[List[Any]] = None,
        pipeline: Optional[Pipeline] = None,
    ):
        if pipeline is None:
            pipelines = cls._get_all()
            for pln in pipelines:
                cls.__add_subscriber(callback, params, pln)
            return

        cls.__add_subscriber(callback, params, pipeline)

    @classmethod
    def _unsubscribe(
        cls,
        callback: Callable[[Pipeline, Job], None],
        params: Optional[List[Any]] = None,
        pipeline: Optional[Pipeline] = None,
    ):
        if pipeline is None:
            pipelines = cls._get_all()
            for pln in pipelines:
                cls.__remove_subscriber(callback, params, pln)
            return

        cls.__remove_subscriber(callback, params, pipeline)

    @classmethod
    def __add_subscriber(cls, callback, params, pipeline):
        pipeline._add_subscriber(callback, params)
        _publish_event(cls._EVENT_ENTITY_TYPE, pipeline.id, EventOperation.UPDATE, "subscribers")

    @classmethod
    def __remove_subscriber(cls, callback, params, pipeline):
        pipeline._remove_subscriber(callback, params)
        _publish_event(cls._EVENT_ENTITY_TYPE, pipeline.id, EventOperation.UPDATE, "subscribers")

    @classmethod
    def _get_or_create(
        cls,
        pipeline_config: PipelineConfig,
        cycle_id: Optional[CycleId] = None,
        scenario_id: Optional[ScenarioId] = None,
    ) -> Pipeline:
        pipeline_id = Pipeline._new_id(str(pipeline_config.id))

        task_configs = [Config.tasks[t.id] for t in pipeline_config.task_configs]
        task_manager = _TaskManagerFactory._build_manager()
        tasks = task_manager._bulk_get_or_create(task_configs, cycle_id, scenario_id, pipeline_id)

        scope = min(task.scope for task in tasks) if len(tasks) != 0 else Scope.GLOBAL
        owner_id: Union[Optional[PipelineId], Optional[ScenarioId], Optional[CycleId]]
        if scope == Scope.PIPELINE:
            owner_id = pipeline_id
        elif scope == Scope.SCENARIO:
            owner_id = scenario_id
        elif scope == Scope.CYCLE:
            owner_id = cycle_id
        else:
            owner_id = None

        if pipelines_from_owner := cls._repository._get_by_config_and_owner_id(str(pipeline_config.id), owner_id):
            return pipelines_from_owner

        version = _VersionManagerFactory._build_manager()._get_latest_version()
        pipeline = Pipeline(
            str(pipeline_config.id),
            dict(**pipeline_config._properties),
            tasks,
            pipeline_id,
            owner_id,
            {scenario_id} if scenario_id else None,
            version=version,
        )
        for task in tasks:
            task._parent_ids.update([pipeline_id])
        cls.__save_tasks(tasks)
        cls._set(pipeline)
        _publish_event(cls._EVENT_ENTITY_TYPE, pipeline.id, EventOperation.CREATION, None)
        return pipeline

    @classmethod
    def __save_tasks(cls, tasks):
        task_manager = _TaskManagerFactory._build_manager()
        for i in tasks:
            task_manager._set(i)

    @classmethod
    def _submit(
        cls,
        pipeline: Union[PipelineId, Pipeline],
        callbacks: Optional[List[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
        check_inputs_are_ready: bool = True,
    ) -> List[Job]:
        pipeline_id = pipeline.id if isinstance(pipeline, Pipeline) else pipeline
        pipeline = cls._get(pipeline_id)
        if pipeline is None:
            raise NonExistingPipeline(pipeline_id)
        callbacks = callbacks or []
        pipeline_subscription_callback = cls.__get_status_notifier_callbacks(pipeline) + callbacks
        if check_inputs_are_ready:
            _warn_if_inputs_not_ready(pipeline._get_inputs())

        jobs = (
            _TaskManagerFactory._build_manager()
            ._orchestrator()
            .submit(pipeline, callbacks=pipeline_subscription_callback, force=force, wait=wait, timeout=timeout)
        )
        _publish_event(cls._EVENT_ENTITY_TYPE, pipeline.id, EventOperation.SUBMISSION, None)
        return jobs

    @staticmethod
    def __get_status_notifier_callbacks(pipeline: Pipeline) -> List:
        return [partial(c.callback, *c.params, pipeline) for c in pipeline.subscribers]

    @classmethod
    def _hard_delete(cls, pipeline_id: PipelineId):
        pipeline = cls._get(pipeline_id)
        entity_ids_to_delete = cls._get_children_entity_ids(pipeline)
        entity_ids_to_delete.pipeline_ids.add(pipeline.id)
        cls._delete_entities_of_multiple_types(entity_ids_to_delete)

    @classmethod
    def _get_children_entity_ids(cls, pipeline: Pipeline) -> _EntityIds:
        entity_ids = _EntityIds()
        for task in pipeline.tasks.values():
            if not isinstance(task, Task):
                task = _TaskManagerFactory._build_manager()._get(task)
            if task.owner_id == pipeline.id:
                entity_ids.task_ids.add(task.id)
            for data_node in task.data_nodes.values():
                if data_node.owner_id == pipeline.id:
                    entity_ids.data_node_ids.add(data_node.id)
        jobs = _JobManagerFactory._build_manager()._get_all()
        for job in jobs:
            if job.task.id in entity_ids.task_ids:
                entity_ids.job_ids.add(job.id)
        return entity_ids



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from queue import SimpleQueue
from typing import Dict, Optional, Set, Tuple

from .event import Event, EventEntityType, EventOperation
from .registration import Registration
from .topic import Topic


def _publish_event(
    entity_type: EventEntityType,
    entity_id: Optional[str],
    operation: EventOperation,
    attribute_name: Optional[str] = None,
):
    Notifier.publish(Event(entity_type, entity_id, operation, attribute_name))


class Notifier:
    _topics_registrations_list: Dict[Topic, Set[Registration]] = {}

    @classmethod
    def register(
        cls,
        entity_type: Optional[EventEntityType] = None,
        entity_id: Optional[str] = None,
        operation: Optional[EventOperation] = None,
        attribute_name: Optional[str] = None,
    ) -> Tuple[str, SimpleQueue]:
        registration = Registration(entity_type, entity_id, operation, attribute_name)

        if registrations := cls._topics_registrations_list.get(registration.topic, None):
            registrations.add(registration)
        else:
            cls._topics_registrations_list[registration.topic] = {registration}

        return registration.registration_id, registration.queue

    @classmethod
    def unregister(cls, registration_id: str):
        to_remove_registration: Optional[Registration] = None

        for _, registrations in cls._topics_registrations_list.items():
            for registration in registrations:
                if registration.registration_id == registration_id:
                    to_remove_registration = registration
                    break

        if to_remove_registration:
            registrations = cls._topics_registrations_list[to_remove_registration.topic]
            registrations.remove(to_remove_registration)
            if len(registrations) == 0:
                del cls._topics_registrations_list[to_remove_registration.topic]

    @classmethod
    def publish(cls, event):
        for topic, registrations in cls._topics_registrations_list.items():
            if Notifier.is_matching(event, topic):
                for registration in registrations:
                    registration.queue.put(event)

    @staticmethod
    def is_matching(event: Event, topic: Topic) -> bool:
        if topic.entity_type is not None and event.entity_type != topic.entity_type:
            return False
        if topic.entity_id is not None and event.entity_id != topic.entity_id:
            return False
        if topic.operation is not None and event.operation != topic.operation:
            return False
        if topic.attribute_name is not None and event.attribute_name and event.attribute_name != topic.attribute_name:
            return False
        return True



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Iterable

from taipy.logger._taipy_logger import _TaipyLogger

from ..data import DataNode


def _warn_if_inputs_not_ready(inputs: Iterable[DataNode]):
    from ..data import CSVDataNode, ExcelDataNode, JSONDataNode, ParquetDataNode, PickleDataNode
    from ..data._data_manager_factory import _DataManagerFactory

    logger = _TaipyLogger._get_logger()
    data_manager = _DataManagerFactory._build_manager()
    for dn in inputs:
        dn = data_manager._get(dn.id)
        if dn.is_ready_for_reading is False and not dn._last_edit_date:
            if dn.storage_type() in [
                CSVDataNode.storage_type(),
                ExcelDataNode.storage_type(),
                JSONDataNode.storage_type(),
                PickleDataNode.storage_type(),
                ParquetDataNode.storage_type(),
            ]:
                logger.warning(
                    f"{dn.id} cannot be read because it has never been written. "
                    f"Hint: The data node may refer to a wrong path : {dn.path} "
                )
            else:
                logger.warning(f"{dn.id} cannot be read because it has never been written.")



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import uuid
from typing import List, Optional, Union

from taipy.config import Config
from taipy.config._config_comparator._comparator_result import _ComparatorResult
from taipy.config.exceptions.exceptions import InconsistentEnvVariableError
from taipy.logger._taipy_logger import _TaipyLogger

from .._manager._manager import _Manager
from ..exceptions.exceptions import ModelNotFound, NonExistingVersion
from ._version import _Version
from ._version_repository_factory import _VersionRepositoryFactory


class _VersionManager(_Manager[_Version]):
    _ENTITY_NAME = _Version.__name__

    __logger = _TaipyLogger._get_logger()

    __DEVELOPMENT_VERSION = ["development", "dev"]
    __LATEST_VERSION = "latest"
    __PRODUCTION_VERSION = "production"
    __ALL_VERSION = ["all", ""]

    __DEFAULT_VERSION = __LATEST_VERSION

    _repository = _VersionRepositoryFactory._build_repository()

    @classmethod
    def _get(cls, entity: Union[str, _Version], default=None) -> _Version:
        """
        Returns the version entity by id or reference.
        """
        entity_id = entity if isinstance(entity, str) else entity.id
        try:
            return cls._repository.load(entity_id)
        except ModelNotFound:
            return default

    @classmethod
    def _get_or_create(cls, id: str, force: bool) -> _Version:
        if version := cls._get(id):
            comparator_result = Config._comparator._find_conflict_config(version.config, Config._applied_config, id)
            if comparator_result.get(_ComparatorResult.CONFLICTED_SECTION_KEY):
                if force:
                    _TaipyLogger._get_logger().warning(f"Overriding version {id} ...")
                    version.config = Config._applied_config
                else:
                    raise SystemExit("The application is stopped. Please check the error log for more information.")
        else:
            version = _Version(id=id, config=Config._applied_config)

        cls._set(version)
        return version

    @classmethod
    def _get_all(cls, version_number: Optional[str] = "all") -> List[_Version]:
        """
        Returns all entities.
        """
        return cls._repository._load_all(version_number)

    @classmethod
    def _get_all_by(cls, by, version_number: Optional[str] = "all") -> List[_Version]:
        """
        Returns all entities based on a criteria.
        """
        return cls._repository._load_all_by(by, version_number)

    @classmethod
    def _set_development_version(cls, version_number: str) -> str:
        cls._get_or_create(version_number, force=True)
        cls._repository._set_development_version(version_number)
        return version_number

    @classmethod
    def _get_development_version(cls) -> str:
        try:
            return cls._repository._get_development_version()
        except (FileNotFoundError, ModelNotFound):
            return cls._set_development_version(str(uuid.uuid4()))

    @classmethod
    def _set_experiment_version(cls, version_number: str, force: bool = False) -> str:
        if version_number == cls._get_development_version():
            raise SystemExit(
                f"Version number {version_number} is already a development version. Please choose a different version "
                f"number for experiment mode. "
            )

        if version_number in cls._get_production_versions():
            raise SystemExit(
                f"Version number {version_number} is already a production version. Please choose a different version "
                f"number for experiment mode. "
            )

        cls._get_or_create(version_number, force)
        cls._repository._set_latest_version(version_number)
        return version_number

    @classmethod
    def _get_latest_version(cls) -> str:
        try:
            return cls._repository._get_latest_version()
        except (FileNotFoundError, ModelNotFound):
            # If there is no version in the system yet, create a new version as development version
            # This set the default versioning behavior on Jupyter notebook to Development mode
            return cls._set_development_version(str(uuid.uuid4()))

    @classmethod
    def _set_production_version(cls, version_number: str, force: bool = False) -> str:
        production_versions = cls._get_production_versions()

        # Check if all previous production versions are compatible with current Python Config
        for production_version in production_versions:
            if production_version == version_number:
                continue

            if version := cls._get(production_version):
                comparator_result = Config._comparator._find_conflict_config(
                    version.config, Config._applied_config, production_version
                )
                if comparator_result.get(_ComparatorResult.CONFLICTED_SECTION_KEY):
                    raise SystemExit("The application is stopped. Please check the error log for more information.")

            else:
                raise NonExistingVersion(production_version)

        if version_number == cls._get_development_version():
            cls._set_development_version(str(uuid.uuid4()))

        cls._get_or_create(version_number, force)
        cls._repository._set_production_version(version_number)
        return version_number

    @classmethod
    def _get_production_versions(cls) -> List[str]:
        try:
            return cls._repository._get_production_versions()
        except (FileNotFoundError, ModelNotFound):
            return []

    @classmethod
    def _delete_production_version(cls, version_number) -> str:
        return cls._repository._delete_production_version(version_number)

    @classmethod
    def _replace_version_number(cls, version_number: Optional[str]):
        if version_number is None:
            version_number = cls._replace_version_number(cls.__DEFAULT_VERSION)

            production_versions = cls._get_production_versions()

            if version_number not in production_versions:
                return version_number
            return production_versions

        if version_number == cls.__LATEST_VERSION:
            return cls._get_latest_version()
        if version_number in cls.__DEVELOPMENT_VERSION:
            return cls._get_development_version()
        if version_number in cls.__PRODUCTION_VERSION:
            return cls._get_production_versions()
        if version_number in cls.__ALL_VERSION:
            return ""

        try:
            if version := cls._get(version_number):
                return version.id
        except InconsistentEnvVariableError:  # The version exist but the Config is alternated
            return version_number

        raise NonExistingVersion(version_number)

    @classmethod
    def _delete_entities_of_multiple_types(cls, _entity_ids):
        return NotImplementedError

    @classmethod
    def _manage_version(cls):
        from ..taipy import clean_all_entities_by_version

        if Config.core.mode == "development":
            current_version_number = cls._get_development_version()
            cls.__logger.info(f"Development mode: Clean all entities of version {current_version_number}")
            clean_all_entities_by_version(current_version_number)
            cls._set_development_version(current_version_number)

        elif Config.core.mode in ["experiment", "production"]:
            default_version_number = {
                "experiment": str(uuid.uuid4()),
                "production": cls._get_latest_version(),
            }
            version_setter = {
                "experiment": cls._set_experiment_version,
                "production": cls._set_production_version,
            }
            if Config.core.version_number:
                current_version_number = Config.core.version_number
            else:
                current_version_number = default_version_number[Config.core.mode]
            if Config.core.clean_entities:
                if clean_all_entities_by_version(current_version_number):
                    cls.__logger.info(f"Clean all entities of version {current_version_number}")
            version_setter[Config.core.mode](current_version_number, Config.core.force)
        else:
            raise SystemExit(f"Undefined execution mode: {Config.core.mode}.")



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
from __future__ import annotations

import abc
from typing import Any, Callable, List, Optional, Set, Union

import networkx as nx

from ..common._listattributes import _ListAttributes
from ..common._utils import _Subscriber
from ..data.data_node import DataNode
from ..job.job import Job
from ..task.task import Task
from ._dag import _DAG


class _Submittable:
    def __init__(self, subscribers=None):
        self._subscribers = _ListAttributes(self, subscribers or list())

    @abc.abstractmethod
    def submit(
        self,
        callbacks: Optional[List[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
    ):
        raise NotImplementedError

    def _get_inputs(self) -> Set[DataNode]:
        dag = self._build_dag()
        return {node for node, degree in dict(dag.in_degree).items() if degree == 0 and isinstance(node, DataNode)}

    @abc.abstractmethod
    def subscribe(self, callback: Callable[[_Submittable, Job], None], params: Optional[List[Any]] = None):
        raise NotImplementedError

    @abc.abstractmethod
    def unsubscribe(self, callback: Callable[[_Submittable, Job], None], params: Optional[List[Any]] = None):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_set_of_tasks(self) -> Set[Task]:
        raise NotImplementedError

    def _get_dag(self) -> _DAG:
        return _DAG(self._build_dag())

    def _build_dag(self) -> nx.DiGraph:
        graph = nx.DiGraph()
        tasks = self._get_set_of_tasks()
        for task in tasks:
            if has_input := task.input:
                for predecessor in task.input.values():
                    graph.add_edges_from([(predecessor, task)])
            if has_output := task.output:
                for successor in task.output.values():
                    graph.add_edges_from([(task, successor)])
            if not has_input and not has_output:
                graph.add_node(task)
        return graph

    def _get_sorted_tasks(self) -> List[List[Task]]:
        dag = self._build_dag()
        remove = [node for node, degree in dict(dag.in_degree).items() if degree == 0 and isinstance(node, DataNode)]
        dag.remove_nodes_from(remove)
        return list(nodes for nodes in nx.topological_generations(dag) if (Task in (type(node) for node in nodes)))

    def _add_subscriber(self, callback: Callable, params: Optional[List[Any]] = None):
        params = [] if params is None else params
        self._subscribers.append(_Subscriber(callback=callback, params=params))

    def _remove_subscriber(self, callback: Callable, params: Optional[List[Any]] = None):
        if params is not None:
            self._subscribers.remove(_Subscriber(callback, params))
        else:
            elem = [x for x in self._subscribers if x.callback == callback]
            if not elem:
                raise ValueError
            self._subscribers.remove(elem[0])



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
import threading
from queue import Empty, SimpleQueue

from .event import Event


class CoreEventConsumerBase(threading.Thread):
    """
    Abstract class showing an example on how to implement a Core event consumer.

    It can be used as a parent class so only the business logic has to be implemented in process_event method.

    """

    def __init__(self, registration_id: str, queue: SimpleQueue):
        threading.Thread.__init__(self, name=f"Thread-Taipy-Core-Consumer-{registration_id}")
        self.daemon = True
        self.queue = queue
        self.__STOP_FLAG = False

    def start(self):
        self.__STOP_FLAG = False
        threading.Thread.start(self)

    def stop(self):
        self.__STOP_FLAG = True

    def run(self):
        while not self.__STOP_FLAG:
            try:
                event: Event = self.queue.get(block=True, timeout=1)
                self.process_event(event)
            except Empty:
                pass

    @abc.abstractmethod
    def process_event(self, event: Event):
        raise NotImplementedError



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import pickle
from datetime import datetime, timedelta
from typing import Any, List, Optional, Set

import modin.pandas as pd

from taipy.config.common.scope import Scope

from .._backup._backup import _replace_in_backup_file
from .._entity._reload import _self_reload
from .._version._version_manager_factory import _VersionManagerFactory
from .abstract_file import _AbstractFileDataNode
from .data_node import DataNode
from .data_node_id import DataNodeId, Edit


class PickleDataNode(DataNode, _AbstractFileDataNode):
    """Data Node stored as a pickle file.

    Attributes:
        config_id (str): Identifier of the data node configuration. It must be a valid Python
            identifer.
        scope (Scope^): The scope of this data node.
        id (str): The unique identifier of this data node.
        name (str): A user-readable name of this data node.
        owner_id (str): The identifier of the owner (pipeline_id, scenario_id, cycle_id) or
            `None`.
        parent_ids (Optional[Set[str]]): The identifiers of the parent tasks or `None`.
        last_edit_date (datetime): The date and time of the last modification.
        edits (List[Edit^]): The ordered list of edits for that job.
        version (str): The string indicates the application version of the data node to instantiate. If not provided,
            the current version is used.
        validity_period (Optional[timedelta]): The duration implemented as a timedelta since the last edit date for
            which the data node can be considered up-to-date. Once the validity period has passed, the data node is
            considered stale and relevant tasks will run even if they are skippable (see the
            [Task management page](../core/entities/task-mgt.md) for more details).
            If _validity_period_ is set to `None`, the data node is always up-to-date.
        edit_in_progress (bool): True if a task computing the data node has been submitted
            and not completed yet. False otherwise.
        properties (dict[str, Any]): A dictionary of additional properties.
            When creating a pickle data node, if the _properties_ dictionary contains a
            _"default_data"_ entry, the data node is automatically written with the corresponding
            _"default_data"_ value.
            If the _properties_ dictionary contains a _"default_path"_ or _"path"_ entry, the data will be stored
            using the corresponding value as the name of the pickle file.
    """

    __STORAGE_TYPE = "pickle"
    __PATH_KEY = "path"
    __DEFAULT_PATH_KEY = "default_path"
    __DEFAULT_DATA_KEY = "default_data"
    __IS_GENERATED_KEY = "is_generated"
    _REQUIRED_PROPERTIES: List[str] = []

    def __init__(
        self,
        config_id: str,
        scope: Scope,
        id: Optional[DataNodeId] = None,
        name: Optional[str] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        last_edit_date: Optional[datetime] = None,
        edits: List[Edit] = None,
        version: str = None,
        validity_period: Optional[timedelta] = None,
        edit_in_progress: bool = False,
        properties=None,
    ):
        if properties is None:
            properties = {}
        default_value = properties.pop(self.__DEFAULT_DATA_KEY, None)
        self._path = properties.get(self.__PATH_KEY, properties.get(self.__DEFAULT_PATH_KEY))
        if self._path is not None:
            properties[self.__PATH_KEY] = self._path
        self._is_generated = properties.get(self.__IS_GENERATED_KEY, self._path is None)
        properties[self.__IS_GENERATED_KEY] = self._is_generated
        super().__init__(
            config_id,
            scope,
            id,
            name,
            owner_id,
            parent_ids,
            last_edit_date,
            edits,
            version or _VersionManagerFactory._build_manager()._get_latest_version(),
            validity_period,
            edit_in_progress,
            **properties,
        )
        if self._path is None:
            self._path = self._build_path(self.storage_type())
        if not self._last_edit_date and os.path.exists(self._path):
            self.last_edit_date = datetime.now()  # type: ignore
        if default_value is not None and not os.path.exists(self._path):
            self.write(default_value)

    @classmethod
    def storage_type(cls) -> str:
        return cls.__STORAGE_TYPE

    @property  # type: ignore
    @_self_reload(DataNode._MANAGER_NAME)
    def path(self) -> Any:
        return self._path

    @path.setter  # type: ignore
    def path(self, value):
        tmp_old_path = self._path
        self._path = value
        self.properties[self.__PATH_KEY] = value
        self.properties[self.__IS_GENERATED_KEY] = False
        _replace_in_backup_file(old_file_path=tmp_old_path, new_file_path=self._path)

    @property  # type: ignore
    @_self_reload(DataNode._MANAGER_NAME)
    def is_generated(self) -> bool:
        return self._is_generated

    def _read(self):
        os.environ["MODIN_PERSISTENT_PICKLE"] = "True"
        with open(self._path, "rb") as pf:
            return pickle.load(pf)

    def _write(self, data):
        if isinstance(data, (pd.DataFrame, pd.Series)):
            os.environ["MODIN_PERSISTENT_PICKLE"] = "True"
        with open(self._path, "wb") as pf:
            pickle.dump(data, pf)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


import os

from taipy import Config

__BACKUP_FILE_PATH_ENVIRONMENT_VARIABLE_NAME = "TAIPY_BACKUP_FILE_PATH"


def _init_backup_file_with_storage_folder():
    if preserve_file_path := os.getenv(__BACKUP_FILE_PATH_ENVIRONMENT_VARIABLE_NAME):
        with open(preserve_file_path, "a") as f:
            f.write(f"{Config.global_config.storage_folder}\n")


def _append_to_backup_file(new_file_path: str):
    if preserve_file_path := os.getenv(__BACKUP_FILE_PATH_ENVIRONMENT_VARIABLE_NAME):
        storage_folder = os.path.abspath(Config.global_config.storage_folder) + os.sep
        if not os.path.abspath(new_file_path).startswith(storage_folder):
            with open(preserve_file_path, "a") as f:
                f.write(f"{new_file_path}\n")


def _remove_from_backup_file(to_remove_file_path: str):
    if preserve_file_path := os.getenv(__BACKUP_FILE_PATH_ENVIRONMENT_VARIABLE_NAME, None):
        storage_folder = os.path.abspath(Config.global_config.storage_folder) + os.sep
        if not os.path.abspath(to_remove_file_path).startswith(storage_folder):
            try:
                with open(preserve_file_path, "r+") as f:
                    old_backup = f.read()
                    to_remove_file_path = to_remove_file_path + "\n"

                    # To avoid removing the file path of different data nodes that are pointing
                    # to the same file. We will only replace the file path only once.
                    if old_backup.startswith(to_remove_file_path):
                        new_backup = old_backup.replace(to_remove_file_path, "", 1)
                    else:
                        new_backup = old_backup.replace("\n" + to_remove_file_path, "\n", 1)

                    if new_backup is not old_backup:
                        f.seek(0)
                        f.write(new_backup)
                        f.truncate()
            except Exception:
                pass


def _replace_in_backup_file(old_file_path: str, new_file_path: str):
    _remove_from_backup_file(old_file_path)
    _append_to_backup_file(new_file_path)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import pathlib
from typing import Optional

from ._job_model import _JobModel
from ._job_repository import _JobRepository


class _JobFSRepository(_JobRepository):
    def __init__(self):
        super().__init__(model=_JobModel, dir_name="jobs")

    @property
    def dir_path(self):
        return super().repo.dir_path

    @property
    def _storage_folder(self) -> pathlib.Path:
        return super().repo._storage_folder

    def _get_by_config_and_owner_id(self, config_id: str, owner_id: Optional[str]):
        return super().repo._get_by_config_and_owner_id(config_id, owner_id)

    def _get_by_configs_and_owner_ids(self, configs_and_owner_ids):
        return super().repo._get_by_configs_and_owner_ids(configs_and_owner_ids)

    def __match_file_and_get_entity(self, filepath, config_and_owner_ids, retry: Optional[int] = 0):
        return super().repo.__match_file_and_get_entity(filepath, config_and_owner_ids, retry)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Optional

from taipy.config import Config
from taipy.logger._taipy_logger import _TaipyLogger

from ._backup._backup import _init_backup_file_with_storage_folder
from ._core_cli import _CoreCLI
from ._orchestrator._dispatcher._job_dispatcher import _JobDispatcher
from ._orchestrator._orchestrator import _Orchestrator
from ._orchestrator._orchestrator_factory import _OrchestratorFactory
from ._version._version_manager_factory import _VersionManagerFactory
from .config import CoreSection


class Core:
    """
    Core service
    """

    __logger = _TaipyLogger._get_logger()

    _orchestrator: Optional[_Orchestrator] = None
    _dispatcher: Optional[_JobDispatcher] = None

    def __init__(self):
        """
        Initialize a Core service.
        """
        self._orchestrator = _OrchestratorFactory._build_orchestrator()

    def run(self, force_restart=False):
        """
        Start a Core service.

        This function checks the configuration, manages application's version,
        and starts a dispatcher and lock the Config.
        """
        self.__update_and_check_config()
        self.__manage_version()
        self.__start_dispatcher(force_restart)

    def stop(self):
        """
        Stop the Core service.

        This function stops the dispatcher and unblock the Config for update.
        """
        Config.unblock_update()

        if self._dispatcher:
            self._dispatcher = _OrchestratorFactory._remove_dispatcher()
            self.__logger.info("Core service has been stopped.")

    @staticmethod
    def __update_and_check_config():
        _CoreCLI.create_parser()
        Config._applied_config._unique_sections[CoreSection.name]._update(_CoreCLI.parse_arguments())
        Config.check()
        Config.block_update()
        _init_backup_file_with_storage_folder()

    @staticmethod
    def __manage_version():
        _VersionManagerFactory._build_manager()._manage_version()

    def __start_dispatcher(self, force_restart):
        if dispatcher := _OrchestratorFactory._build_dispatcher(force_restart=force_restart):
            self._dispatcher = dispatcher

        if Config.job_config.is_development:
            _Orchestrator._check_and_execute_jobs_if_development_mode()



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from taipy.config._config import _Config
from taipy.config.checker._checkers._config_checker import _ConfigChecker
from taipy.config.checker.issue_collector import IssueCollector
from taipy.config.common.frequency import Frequency

from ..pipeline_config import PipelineConfig
from ..scenario_config import ScenarioConfig


class _ScenarioConfigChecker(_ConfigChecker):
    def __init__(self, config: _Config, collector: IssueCollector):
        super().__init__(config, collector)

    def _check(self) -> IssueCollector:
        scenario_configs = self._config._sections[ScenarioConfig.name]
        for scenario_config_id, scenario_config in scenario_configs.items():
            if scenario_config_id != _Config.DEFAULT_KEY:
                self._check_if_entity_property_key_used_is_predefined(scenario_config)
                self._check_existing_config_id(scenario_config)
                self._check_frequency(scenario_config_id, scenario_config)
                self._check_pipelines(scenario_config_id, scenario_config)
                self._check_comparators(scenario_config_id, scenario_config)
        return self._collector

    def _check_pipelines(self, scenario_config_id: str, scenario_config: ScenarioConfig):
        self._check_children(
            ScenarioConfig,
            scenario_config_id,
            scenario_config._PIPELINE_KEY,
            scenario_config.pipeline_configs,
            PipelineConfig,
        )

    def _check_frequency(self, scenario_config_id: str, scenario_config: ScenarioConfig):
        if scenario_config.frequency and not isinstance(scenario_config.frequency, Frequency):
            self._error(
                scenario_config._FREQUENCY_KEY,
                scenario_config.frequency,
                f"{scenario_config._FREQUENCY_KEY} field of ScenarioConfig `{scenario_config_id}` must be"
                f" populated with a Frequency value.",
            )

    def _check_comparators(self, scenario_config_id: str, scenario_config: ScenarioConfig):
        if not scenario_config.comparators:
            self._info(
                scenario_config._COMPARATOR_KEY,
                scenario_config.comparators,
                f"No scenario {scenario_config._COMPARATOR_KEY} defined for ScenarioConfig `{scenario_config_id}`.",
            )



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


from typing import Union

from sqlalchemy import Boolean, Column, Integer, Text
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
Json = Union[dict, list, str, int, float, bool, None]


class _TaipyModel(Base):  # type: ignore
    """_TaipyModel Represents the table used to store the information of taipy entities."""

    __tablename__ = "taipy_model"

    id = Column(Integer, primary_key=True)
    model_id = Column(Text)
    model_name = Column(Text)
    document = Column(Text)

    def __int__(self, model_id: str, model_name: str, document: Json):
        self.model_id = model_id
        self.model_name = model_name
        self.document = document


class _TaipyVersion(Base):  # type: ignore
    """_TaipyVersion Represents the table used to store versioning information of taipy configs."""

    __tablename__ = "taipy_version"

    id = Column(Text, primary_key=True)
    config = Column(Text)
    creation_date = Column(Text)
    is_production = Column(Boolean)
    is_development = Column(Boolean)
    is_latest = Column(Boolean)

    def __int__(
        self,
        id: str,
        config: Json,
        creation_date: str,
    ):
        self.id = id
        self.config = config
        self.creation_date = creation_date
        self.is_production = False
        self.is_development = False
        self.latest = False



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
from typing import Dict, Iterable, List, Optional, Set, Union

from taipy.config._config import _Config
from taipy.config.common.scope import Scope
from taipy.config.config import Config

from .._backup._backup import _append_to_backup_file, _remove_from_backup_file
from .._manager._manager import _Manager
from .._version._version_manager_factory import _VersionManagerFactory
from ..config.data_node_config import DataNodeConfig
from ..cycle.cycle_id import CycleId
from ..exceptions.exceptions import InvalidDataNodeType
from ..notification import EventEntityType, EventOperation, _publish_event
from ..pipeline.pipeline_id import PipelineId
from ..scenario.scenario_id import ScenarioId
from ._data_repository_factory import _DataRepositoryFactory
from .abstract_file import _AbstractFileDataNode
from .data_node import DataNode
from .data_node_id import DataNodeId
from .pickle import PickleDataNode


class _DataManager(_Manager[DataNode]):
    __DATA_NODE_CLASS_MAP = DataNode._class_map()  # type: ignore
    _repository = _DataRepositoryFactory._build_repository()  # type: ignore
    _ENTITY_NAME = DataNode.__name__
    _EVENT_ENTITY_TYPE = EventEntityType.DATA_NODE

    @classmethod
    def _bulk_get_or_create(
        cls,
        data_node_configs: List[DataNodeConfig],
        cycle_id: Optional[CycleId] = None,
        scenario_id: Optional[ScenarioId] = None,
        pipeline_id: Optional[PipelineId] = None,
    ) -> Dict[DataNodeConfig, DataNode]:
        data_node_configs = [Config.data_nodes[dnc.id] for dnc in data_node_configs]
        dn_configs_and_owner_id = []
        for dn_config in data_node_configs:
            scope = dn_config.scope
            owner_id: Union[Optional[PipelineId], Optional[ScenarioId], Optional[CycleId]]
            if scope == Scope.PIPELINE:
                owner_id = pipeline_id
            elif scope == Scope.SCENARIO:
                owner_id = scenario_id
            elif scope == Scope.CYCLE:
                owner_id = cycle_id
            else:
                owner_id = None
            dn_configs_and_owner_id.append((dn_config, owner_id))

        data_nodes = cls._repository._get_by_configs_and_owner_ids(dn_configs_and_owner_id)  # type: ignore

        return {
            dn_config: data_nodes.get((dn_config, owner_id)) or cls._create_and_set(dn_config, owner_id, None)
            for dn_config, owner_id in dn_configs_and_owner_id
        }

    @classmethod
    def _create_and_set(
        cls, data_node_config: DataNodeConfig, owner_id: Optional[str], parent_ids: Optional[Set[str]]
    ) -> DataNode:
        data_node = cls.__create(data_node_config, owner_id, parent_ids)
        cls._set(data_node)
        if isinstance(data_node, _AbstractFileDataNode):
            _append_to_backup_file(new_file_path=data_node._path)
        _publish_event(cls._EVENT_ENTITY_TYPE, data_node.id, EventOperation.CREATION, None)
        return data_node

    @classmethod
    def __create(
        cls, data_node_config: DataNodeConfig, owner_id: Optional[str], parent_ids: Optional[Set[str]]
    ) -> DataNode:
        try:
            version = _VersionManagerFactory._build_manager()._get_latest_version()
            props = data_node_config._properties.copy()

            if data_node_config.storage_type:
                storage_type = data_node_config.storage_type
            else:
                storage_type = Config.sections[DataNodeConfig.name][_Config.DEFAULT_KEY].storage_type

            return cls.__DATA_NODE_CLASS_MAP[storage_type](
                config_id=data_node_config.id,
                scope=data_node_config.scope or DataNodeConfig._DEFAULT_SCOPE,
                validity_period=data_node_config.validity_period,
                owner_id=owner_id,
                parent_ids=parent_ids,
                version=version,
                properties=props,
            )
        except KeyError:
            raise InvalidDataNodeType(data_node_config.storage_type)

    @classmethod
    def _clean_pickle_file(cls, data_node: DataNode):
        if not isinstance(data_node, PickleDataNode):
            return
        if data_node.is_generated and os.path.exists(data_node.path):
            os.remove(data_node.path)

    @classmethod
    def _clean_pickle_files(cls, data_nodes: Iterable[DataNode]):
        for data_node in data_nodes:
            cls._clean_pickle_file(data_node)

    @classmethod
    def _remove_dn_file_path_in_backup_file(cls, data_node: DataNode):
        if isinstance(data_node, _AbstractFileDataNode):
            _remove_from_backup_file(to_remove_file_path=data_node.path)

    @classmethod
    def _remove_dn_file_paths_in_backup_file(cls, data_nodes: Iterable[DataNode]):
        for data_node in data_nodes:
            cls._remove_dn_file_path_in_backup_file(data_node)

    @classmethod
    def _delete(cls, data_node_id: DataNodeId):
        data_node = cls._get(data_node_id, None)
        if data_node:
            cls._clean_pickle_file(data_node)
            cls._remove_dn_file_path_in_backup_file(data_node)
        super()._delete(data_node_id)

    @classmethod
    def _delete_many(cls, data_node_ids: Iterable[DataNodeId]):
        data_nodes = []
        for data_node_id in data_node_ids:
            if data_node := cls._get(data_node_id):
                data_nodes.append(data_node)
        cls._clean_pickle_files(data_nodes)
        cls._remove_dn_file_paths_in_backup_file(data_nodes)
        super()._delete_many(data_node_ids)

    @classmethod
    def _delete_all(cls):
        data_nodes = cls._get_all()
        cls._clean_pickle_files(data_nodes)
        cls._remove_dn_file_paths_in_backup_file(data_nodes)
        super()._delete_all()

    @classmethod
    def _delete_by_version(cls, version_number: str):
        data_nodes = cls._get_all(version_number)
        cls._clean_pickle_files(data_nodes)
        cls._remove_dn_file_paths_in_backup_file(data_nodes)
        cls._repository._delete_by(attribute="version", value=version_number, version_number="all")
        _publish_event(cls._EVENT_ENTITY_TYPE, None, EventOperation.DELETION, None)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from importlib import util
from typing import Optional, Type

from taipy.config.config import Config

from ..common._utils import _load_fct
from ..exceptions.exceptions import ModeNotAvailable, OrchestratorNotBuilt
from ._abstract_orchestrator import _AbstractOrchestrator
from ._dispatcher import _DevelopmentJobDispatcher, _JobDispatcher, _StandaloneJobDispatcher
from ._orchestrator import _Orchestrator


class _OrchestratorFactory:
    _TAIPY_ENTERPRISE_MODULE = "taipy.enterprise"
    _TAIPY_ENTERPRISE_CORE_ORCHESTRATOR_MODULE = _TAIPY_ENTERPRISE_MODULE + ".core._orchestrator._orchestrator"
    _TAIPY_ENTERPRISE_CORE_DISPATCHER_MODULE = _TAIPY_ENTERPRISE_MODULE + ".core._orchestrator._dispatcher"
    __STANDADLONE_JOB_DISPATCHER_TYPE = "_StandaloneJobDispatcher"

    _orchestrator: Optional[_Orchestrator] = None
    _dispatcher: Optional[_JobDispatcher] = None

    @classmethod
    def _build_orchestrator(cls) -> Type[_AbstractOrchestrator]:
        if util.find_spec(cls._TAIPY_ENTERPRISE_MODULE) is not None:
            cls._orchestrator = _load_fct(
                cls._TAIPY_ENTERPRISE_CORE_ORCHESTRATOR_MODULE,
                "Orchestrator",
            )  # type: ignore
        else:
            cls._orchestrator = _Orchestrator  # type: ignore

        cls._orchestrator.initialize()  # type: ignore

        return cls._orchestrator  # type: ignore

    @classmethod
    def _build_dispatcher(cls, force_restart=False) -> Optional[_JobDispatcher]:
        if not cls._orchestrator:
            raise OrchestratorNotBuilt
        if Config.job_config.is_standalone:
            return cls.__build_standalone_job_dispatcher(force_restart=force_restart)
        elif Config.job_config.is_development:
            return cls.__build_development_job_dispatcher()
        else:
            raise ModeNotAvailable(f"Job mode {Config.job_config.mode} is not available.")

    @classmethod
    def _remove_dispatcher(cls) -> Optional[_JobDispatcher]:
        if isinstance(cls._dispatcher, _StandaloneJobDispatcher) and not isinstance(
            cls._dispatcher, _DevelopmentJobDispatcher
        ):
            cls._dispatcher.stop()

        cls._dispatcher = None

        return cls._dispatcher

    @classmethod
    def __build_standalone_job_dispatcher(cls, force_restart=False) -> Optional[_JobDispatcher]:
        if isinstance(cls._dispatcher, _StandaloneJobDispatcher) and not isinstance(
            cls._dispatcher, _DevelopmentJobDispatcher
        ):
            if force_restart:
                cls._dispatcher.stop()
            else:
                return None

        if util.find_spec(cls._TAIPY_ENTERPRISE_MODULE) is not None:
            cls._dispatcher = _load_fct(
                cls._TAIPY_ENTERPRISE_CORE_DISPATCHER_MODULE, cls.__STANDADLONE_JOB_DISPATCHER_TYPE
            )(cls._orchestrator)
        else:
            cls._dispatcher = _StandaloneJobDispatcher(cls._orchestrator)  # type: ignore

        cls._dispatcher.start()  # type: ignore
        return cls._dispatcher

    @classmethod
    def __build_development_job_dispatcher(cls) -> _JobDispatcher:
        if isinstance(cls._dispatcher, _StandaloneJobDispatcher) and not isinstance(
            cls._dispatcher, _DevelopmentJobDispatcher
        ):
            cls._dispatcher.stop()
        cls._dispatcher = _DevelopmentJobDispatcher(cls._orchestrator)  # type: ignore
        return cls._dispatcher



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import math
from functools import reduce
from typing import Any, Dict, List, Tuple

import networkx as nx


class _Node:
    def __init__(self, entity: Any, x, y):
        self.type = entity.__class__.__name__
        self.entity = entity
        self.x = x
        self.y = y


class _Edge:
    def __init__(self, src: _Node, dest: _Node):
        self.src = src
        self.dest = dest


class _DAG:
    def __init__(self, dag: nx.DiGraph):
        self._sorted_nodes = list(nodes for nodes in nx.topological_generations(dag))
        self._length, self._width = self.__compute_size()
        self._grid_length, self._grid_width = self.__compute_grid_size()
        self._nodes = self.__compute_nodes()
        self._edges = self.__compute_edges(dag)

    @property
    def width(self) -> int:
        return self._width

    @property
    def length(self) -> int:
        return self._length

    @property
    def nodes(self) -> Dict[str, _Node]:
        return self._nodes

    @property
    def edges(self) -> List[_Edge]:
        return self._edges

    def __compute_size(self) -> Tuple[int, int]:
        return len(self._sorted_nodes), max([len(i) for i in self._sorted_nodes])

    def __compute_grid_size(self) -> Tuple[int, int]:
        if self._width == 1:
            grd_wdt = 1
        else:
            grd_wdt = self.__lcm(*[len(i) + 1 if len(i) != self._width else len(i) - 1 for i in self._sorted_nodes]) + 1
        return len(self._sorted_nodes), grd_wdt

    def __compute_nodes(self) -> Dict[str, _Node]:
        nodes = {}
        x = 0
        for same_lvl_nodes in self._sorted_nodes:
            lcl_wdt = len(same_lvl_nodes)
            is_max = lcl_wdt != self.width
            if self.width != 1:
                y_incr = (self._grid_width - 1) / (lcl_wdt + 1) if is_max else (self._grid_width - 1) / (lcl_wdt - 1)
            else:
                y_incr = 1
            y = 0 if is_max else -y_incr
            for node in same_lvl_nodes:
                y += y_incr
                nodes[node.id] = _Node(node, x, y)
            x += 1
        return nodes

    def __compute_edges(self, dag) -> List[_Edge]:
        edges = []
        for edge in dag.edges():
            edges.append(_Edge(self.nodes[edge[0].id], self.nodes[edge[1].id]))
        return edges

    @staticmethod
    def __lcm(*integers) -> int:
        # Function math.lcm is only implemented for Python 3.9+
        # For compatibility with Python 3.8 it has been re implemented.
        if 0 in integers:
            return 0
        return reduce(lambda x, y: (x * y) // math.gcd(x, y), integers)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import pathlib
from datetime import datetime
from typing import Any, Iterable, List, Optional, Union

from taipy.config import Config

from .._repository._repository import _AbstractRepository
from .._repository._repository_adapter import _RepositoryAdapter
from ._version import _Version
from ._version_model import _VersionModel


class _VersionRepository(_AbstractRepository[_VersionModel, _Version]):  # type: ignore
    _LATEST_VERSION_KEY = "latest_version"
    _DEVELOPMENT_VERSION_KEY = "development_version"
    _PRODUCTION_VERSION_KEY = "production_version"

    def __init__(self, **kwargs):
        kwargs.update({"to_model_fct": self._to_model, "from_model_fct": self._from_model})
        self.repo = _RepositoryAdapter.select_base_repository()(**kwargs)

    @property
    def repository(self):
        return self.repo

    def _to_model(self, version: _Version):
        return _VersionModel(
            id=version.id, config=Config._to_json(version.config), creation_date=version.creation_date.isoformat()
        )

    def _from_model(self, model):
        version = _Version(id=model.id, config=Config._from_json(model.config))
        version.creation_date = datetime.fromisoformat(model.creation_date)
        return version

    def load(self, model_id: str) -> _Version:
        return self.repo.load(model_id)

    def _load_all(self, version_number: Optional[str] = "all") -> List[_Version]:
        return self.repo._load_all(version_number)

    def _load_all_by(self, by, version_number: Optional[str] = "all") -> List[_Version]:
        return self.repo._load_all_by(by, version_number)

    def _save(self, entity: _Version):
        return self.repo._save(entity)

    def _delete(self, entity_id: str):
        return self.repo._delete(entity_id)

    def _delete_all(self):
        return self.repo._delete_all()

    def _delete_many(self, ids: Iterable[str]):
        return self.repo._delete_many(ids)

    def _delete_by(self, attribute: str, value: str, version_number: Optional[str] = None):
        return self.repo._delete_by(attribute, value, version_number)

    def _search(self, attribute: str, value: Any, version_number: Optional[str] = None) -> Optional[_Version]:
        return self.repo._search(attribute, value, version_number)

    def _export(self, entity_id: str, folder_path: Union[str, pathlib.Path]):
        return self.repo._export(entity_id, folder_path)

    def _set_latest_version(self, version_number):
        self.repo._set_latest_version(version_number)

    def _get_latest_version(self) -> str:
        return self.repo._get_latest_version()

    def _set_development_version(self, version_number):
        self.repo._set_development_version(version_number)

    def _get_development_version(self) -> str:
        return self.repo._get_development_version()

    def _set_production_version(self, version_number):
        self.repo._set_production_version(version_number)

    def _get_production_versions(self) -> List[str]:
        return self.repo._get_production_versions()

    def _delete_production_version(self, version_number):
        self.repo._delete_production_version(version_number)



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import uuid
from datetime import datetime
from typing import Any, Dict, Optional

from taipy.config.common.frequency import Frequency

from .._entity._entity import _Entity
from .._entity._labeled import _Labeled
from .._entity._properties import _Properties
from .._entity._reload import _reload, _self_reload, _self_setter
from ..exceptions.exceptions import _SuspiciousFileOperation
from .cycle_id import CycleId


class Cycle(_Entity, _Labeled):
    """An iteration of a recurrent work pattern.

    Attributes:
        id (str): The unique identifier of the cycle.
        frequency (Frequency^): The frequency of this cycle.
        creation_date (datetime): The date and time of the creation of this cycle.
        start_date (datetime): The date and time of the start of this cycle.
        end_date (datetime): The date and time of the end of this cycle.
        name (str): The name of this cycle.
        properties (dict[str, Any]): A dictionary of additional properties.
    """

    _ID_PREFIX = "CYCLE"
    __SEPARATOR = "_"
    _MANAGER_NAME = "cycle"

    def __init__(
        self,
        frequency: Frequency,
        properties: Dict[str, Any],
        creation_date: datetime,
        start_date: datetime,
        end_date: datetime,
        name: Optional[str] = None,
        id: Optional[CycleId] = None,
    ):
        self._frequency = frequency
        self._creation_date = creation_date
        self._start_date = start_date
        self._end_date = end_date
        self._name = self._new_name(name)
        self.id = id or self._new_id(self._name)
        self._properties = _Properties(self, **properties)

    def _new_name(self, name: Optional[str] = None) -> str:
        if name:
            return name
        if self._frequency == Frequency.DAILY:
            # Example "Monday, 2. January 2023"
            return self._start_date.strftime("%A, %d. %B %Y")
        if self._frequency == Frequency.WEEKLY:
            # Example "Week 01 2023, from 2. January"
            return self._start_date.strftime("Week %W %Y, from %d. %B")
        if self._frequency == Frequency.MONTHLY:
            # Example "January 2023"
            return self._start_date.strftime("%B %Y")
        if self._frequency == Frequency.QUARTERLY:
            # Example "2023 Q1"
            return f"{self._start_date.strftime('%Y')} Q{(self._start_date.month-1)//3+1}"
        if self._frequency == Frequency.YEARLY:
            # Example "2023"
            return self._start_date.strftime("%Y")
        return Cycle.__SEPARATOR.join([str(self._frequency.value), self._start_date.ctime()])

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def frequency(self):
        return self._frequency

    @frequency.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def frequency(self, val):
        self._frequency = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def creation_date(self):
        return self._creation_date

    @creation_date.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def creation_date(self, val):
        self._creation_date = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def start_date(self):
        return self._start_date

    @start_date.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def start_date(self, val):
        self._start_date = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def end_date(self):
        return self._end_date

    @end_date.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def end_date(self, val):
        self._end_date = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def name(self):
        return self._name

    @name.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def name(self, val):
        self._name = val

    @property
    def properties(self):
        self._properties = _reload(self._MANAGER_NAME, self)._properties
        return self._properties

    @staticmethod
    def _new_id(name: str) -> CycleId:
        def _get_valid_filename(name: str) -> str:
            """
            Source: https://github.com/django/django/blob/main/django/utils/text.py
            """
            s = str(name).strip().replace(" ", "_")
            s = re.sub(r"(?u)[^-\w.]", "", s)
            if s in {"", ".", ".."}:
                raise _SuspiciousFileOperation("Could not derive file name from '%s'" % name)
            s = str(s).strip().replace(" ", "_")
            return re.sub(r"(?u)[^-\w.]", "", s)

        return CycleId(_get_valid_filename(Cycle.__SEPARATOR.join([Cycle._ID_PREFIX, name, str(uuid.uuid4())])))

    def __getattr__(self, attribute_name):
        protected_attribute_name = attribute_name
        if protected_attribute_name in self._properties:
            return self._properties[protected_attribute_name]
        raise AttributeError(f"{attribute_name} is not an attribute of cycle {self.id}")

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def get_label(self) -> str:
        """Returns the cycle label.

        Returns:
            The label of the cycle as a string.
        """
        return self._get_label()

    def get_simple_label(self) -> str:
        """Returns the cycle simple label.

        Returns:
            The simple label of the cycle as a string.
        """
        return self._get_simple_label()



# Copyright 2023 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from collections import UserDict

from ..notification import _ENTITY_TO_EVENT_ENTITY_TYPE, EventOperation, _publish_event


class _Properties(UserDict):
    def __init__(self, entity_owner, **kwargs):
        super().__init__(**kwargs)
        self._entity_owner = entity_owner

    def __setitem__(self, key, value):
        super(_Properties, self).__setitem__(key, value)
        from ... import core as tp

        if hasattr(self, "_entity_owner"):
            tp.set(self._entity_owner)
            _publish_event(
                _ENTITY_TO_EVENT_ENTITY_TYPE[self._entity_owner._MANAGER_NAME],
                self._entity_owner.id,
                EventOperation.UPDATE,
                key,
            )

    def __getitem__(self, key):
        from taipy.config.common._template_handler import _TemplateHandler as _tpl

        return _tpl._replace_templates(super(_Properties, self).__getitem__(key))
